/*===========================================================================


       formgrp.cpp -  handler for all objects on a form

                      Copyright  2001, 2003 by Pankhurst Algorithmics
                      (www.pankhurst.com, www.ebmDevMag.com)
                      All Rights Reserved


 ** LICENSE ********************************************************************
 *                                                                             *
 *  Version: MPL 1.1                                                           *
 *                                                                             *
 *  The contents of this file are subject to the Mozilla Public License        *
 *  Version 1.1 (the "License"); you may not use this file except in           *
 *  compliance with the License. You may obtain a copy of the License at       *
 *  http://www.mozilla.org/MPL/                                                *
 *                                                                             *
 *  Software distributed under the License is distributed on an "AS IS" basis, *
 *  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License   *
 *  for the specific language governing rights and limitations under the       *
 *  License.                                                                   *
 *                                                                             *
 *  The Original Code is: eBookMan Form Editor                                 *
 *                                                                             *
 *  The Initial Developer of the Original Code is                              *
 *  Pankhurst Algorithmics                                                     *
 *                                                                             *
 *  Portions created by the Initial Developer are Copyright (C) 2001-2003      *
 *  Pankhurst Algorithmics. All Rights Reserved.                               *
 *                                                                             *
 *                                                                             *
 *******************************************************************************


===========================================================================*/
#include <vcl.h>
#pragma hdrstop

#include "objcreat.h"

#include "formgrp.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
/*===========================================================================


       CFormGroup class - holder/manager for all form objects
       

===========================================================================*/
CFormGroup::CFormGroup(unsigned int maxArray,const char *formName,int x,int y,int dx,int dy)
 : c_maxObj(maxArray),
   m_obj(NULL),
   m_totalObj(0),
   m_dirty(false),
   m_directivesText(NULL),
   m_classText(NULL)
{
  m_directivesText=new char[2]; // start out with something - we'll increase as needed in ReNew()
  m_directivesText[0]='\0';     // and clear it
  m_classText=new char[2];
  m_classText[0]='\0';
  // set up/clear an array of base pointers to our objects
  m_obj=new CBaseObject*[c_maxObj+3];
  for (int i=0;i<c_maxObj;i++)
    m_obj[i]=NULL;
  // by default each class needs a form, so set it up here
  Add(new CObjForm(),x,y,dx,dy);
  m_obj[0]->SetName(formName); // and name it
}
//---------------------------------------------------------------------------
CFormGroup::~CFormGroup(void)
{
  // clean up
  Clear();
  delete[] m_obj;
}
//---------------------------------------------------------------------------
void CFormGroup::Clear(void)
{
  // remove all objects
  for (int i=0;i<c_maxObj;i++)
  {
    delete m_obj[i];
    m_obj[i]=NULL;
  }
  m_totalObj=0;
  m_dirty=false;
}
//---------------------------------------------------------------------------
CBaseObject *CFormGroup::Object(unsigned int index)
{
  // access object pointer, with range checking
  if (index<c_maxObj)
    return m_obj[index];
  return NULL;
}
//---------------------------------------------------------------------------
bool CFormGroup::Add(CBaseObject*object,int x,int y,int dx,int dy)
{
  // add object & set boundaries - child object determines how these values used
  if (!Add(object))
    return false;
  object->SetBounds(x,y,dx,dy);
  m_dirty=true;
  return true;
}
//---------------------------------------------------------------------------
bool CFormGroup::Add(CBaseObject*object)
{
  // add object to list
  if (m_totalObj>=c_maxObj)
  {
    delete object;
    return false;
  }
  m_obj[m_totalObj++]=object;
  m_dirty=true;
  return true;
}
//---------------------------------------------------------------------------
void CFormGroup::Delete(unsigned int index)
{
  // delete object from list & compact
  if (index<c_maxObj)
  {
    delete m_obj[index];
    for (int i=index+1;i<m_totalObj;i++)
      m_obj[i-1]=m_obj[i];
    m_obj[--m_totalObj]=NULL;
    m_dirty=true;
  }
}
//---------------------------------------------------------------------------
void CFormGroup::DefaultHeader(s_fileHeader*hdr,int totalEntry)
{
  // fill header with version info - makes version handling more open
  hdr->versionMain=100;
  hdr->versionType=1;
  hdr->id         =0x19B03EFAL; // for all our products using this file 'style'
  hdr->totalEntry =totalEntry;
}
//---------------------------------------------------------------------------
bool CFormGroup::ReNew(char**ptr,unsigned int oldCopyLen,unsigned int newLen)
{
  // resize character array to fit newLen
  try
  {
    char *temp=new char[newLen+1];
    if ( NULL!=ptr && oldCopyLen>0 )
      memcpy(temp,*ptr,oldCopyLen);
    else
      temp[0]='\0';
    delete[] *ptr;
    *ptr=temp;
    return true;
  }
  catch (...)
  {
    return false;
  }
}
//---------------------------------------------------------------------------
bool CFormGroup::ReadString(HANDLE handle,char **ptrIn)
{
  // get text string in from file - resize destination as needed 
  DWORD bytesProcessed=0;
  unsigned int increment=32, size=increment, k=0;
  char *ptr=new char[size];
  ptr[0]='\0';
  while (1)
  {
    if ( 0==::ReadFile(handle,&ptr[k],1,&bytesProcessed,NULL) || 1!=bytesProcessed )
      break; // done
    if ('\0'==ptr[k])
      break;
    if (++k>size-2)
    {
      // resize with a little extra space so we aren't resizing constantly
      if (!ReNew(&ptr,size,size+increment))
        break;
      size+=increment;
    }
  }
  ptr[k]='\0';
  delete[] *ptrIn;
  *ptrIn=ptr;
  return true;
}
//---------------------------------------------------------------------------
bool CFormGroup::WriteString(HANDLE handle,const char *ptr)
{
  // get string written to disk
  DWORD bytesProcessed=0;
  int len=strlen(ptr)+1;
  if ( 0==WriteFile(handle,ptr,len,&bytesProcessed,NULL) || len!=bytesProcessed )
    return false;
  return true;
}
//---------------------------------------------------------------------------
bool CFormGroup::Save(const char *file)
{
  // save form data to file on disk - write header, object data, string data
  HANDLE handle=INVALID_HANDLE_VALUE;
  try
  {
    handle=CreateFile(file,GENERIC_WRITE|GENERIC_READ,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
    if (INVALID_HANDLE_VALUE==handle)
      throw "Unable to open file";
    DWORD bytesProcessed=0;
    struct s_fileHeader hdr;
    DefaultHeader(&hdr,m_totalObj);
    // write header out
    if ( 0==WriteFile(handle,&hdr,sizeof(hdr),&bytesProcessed,NULL) || sizeof(hdr)!=bytesProcessed )
      throw  "Error writing";
    // write each object's necessary data
    for (int i=0;i<m_totalObj;i++)
    {
      if (0==m_obj[i]->WriteTables(handle))
        throw "Error writing";
    }
    // write source code text strings
    if (!WriteString(handle,m_classText))
      throw "Error reading file";
    if (!WriteString(handle,m_directivesText))
      throw "Error reading file";
    CloseHandle(handle);
    m_dirty=false;
    return true;
  }
  catch (const char *msg) // error handler on failure - display message, clean up
  {
    if (INVALID_HANDLE_VALUE!=handle)
      CloseHandle(handle);
    DeleteFile(file);
    MessageBox(NULL,msg,"Save Error",MB_OK);
    return false;
  }
}
//------------------------------------------------------------------------------
bool CFormGroup::Load(const char *file)
{
  // load form file - do same order as save, in reverse
  HANDLE handle=INVALID_HANDLE_VALUE;
  try
  {
    handle=CreateFile(file,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
    if (INVALID_HANDLE_VALUE==handle)
      throw "Unable to open file";
    DWORD bytesProcessed=0;
    struct s_fileHeader hdr, hdrDefault;
    DefaultHeader(&hdrDefault,0);
    // avoid incompatible version files by checking header
    if ( 0==ReadFile(handle,&hdr,sizeof(hdr),&bytesProcessed,NULL) || sizeof(hdr)!=bytesProcessed )
      throw  "Error reading";
    if (hdr.id         !=hdrDefault.id         )
      throw  "Invalid file";
    if (hdr.versionType!=hdrDefault.versionType)
      throw  "Wrong file type (created with another program)";
    if (hdr.versionMain!=hdrDefault.versionMain)
      throw  "Invalid file version";
    // clear previous entries...
    int i=0, total=min(hdr.totalEntry,(long)c_maxObj);
    Clear();
    CBaseObject *objIn=NULL;
    // now read/create each object in array
    for (int i=0;i<total;i++)
    {
      if ( NULL == (objIn=ReadEntry(handle)) )
        throw "Error reading file";
      Add(objIn);
    }
    // finally, read in source code text strings
    if (!ReadString(handle,&m_classText))
      throw "Error reading file";
    if (!ReadString(handle,&m_directivesText))
      throw "Error reading file";
    CloseHandle(handle);
    m_dirty=false;
    return true;
  }
  catch (const char *msg) // error handler on failure - display message, clean up
  {
    if (INVALID_HANDLE_VALUE!=handle)
      CloseHandle(handle);
    MessageBox(NULL,msg,"Load Error",MB_OK);
    return false;
  }
}
//------------------------------------------------------------------------------
CBaseObject *CFormGroup::ReadEntry(HANDLE hFile)
{
  // each object in the file is identified uniquely, read code and create
  // appropriate object, then return allocated mem/obj
  DWORD bytesProcessed=0, i=0, k=0;
  BYTE temp=0;
  CBaseObject *obj=NULL;
  if ( 0==ReadFile(hFile,&temp,1,&bytesProcessed,NULL) || bytesProcessed!=1 )
    return NULL; // failure (or file end) - either way, unable to continue
  switch (temp)
  {
    case OBJ_FORM:
      obj=new CObjForm();
      break;
    case OBJ_BUTTON:
      obj=new CObjButton();
      break;
    case OBJ_CHECKBOX:
      obj=new CObjCheckBox();
      break;
    case OBJ_TEXTEDIT:
      obj=new CObjTextEdit();
      break;
    case OBJ_SCROLLBAR:
      obj=new CObjScrollBar();
      break;
    case OBJ_LABEL:
      obj=new CObjLabel();
      break;
    case OBJ_TABLE:
      obj=new CObjTable();
      break;
    case OBJ_REPEATBUTTON:
      obj=new CObjRepeatButton();
      break;
    case OBJ_PUSHBUTTON:
      obj=new CObjPushButton();
      break;
    case OBJ_PROGRESSBAR:
      obj=new CObjProgressBar();
      break;
    case OBJ_RECT:
      obj=new CObjRect();
      break;
    case OBJ_ICON:
      obj=new CObjIcon();
      break;
    case OBJ_TIMEEDIT:
      obj=new CObjTimeEdit();
      break;
    case OBJ_DATEEDIT:
      obj=new CObjDateEdit();
      break;
    case OBJ_RADIOBUTTON:
      obj=new CObjRadioButton();
      break;
    default:
      return NULL;
  }
  // once object is created, call it to read in needed (and unique) data
  if (0==obj->ReadTables(hFile))
  {
    delete obj;
    return NULL;
  }
  return obj;
}
//---------------------------------------------------------------------------
bool CFormGroup::OutputCPPSourceCode(const char *fileIn,
                                     bool unixLinefeeds,
                                     bool excludeMain )
{
  // output source - if file name NULL, then place on clipboard - return success flag
  char *buff=new char[ 64000 ]; // should be enough!
  bool success=false;
  if (NULL!=fileIn) // writing to a file?
  {
    char headerFile[260]="", *ptr=NULL, fileId[260]="", file[260]="";
    // massage string, and make sure with have one with .h at end and other
    // with .c or .cpp - then write it
    strcpy(headerFile,fileIn);
    if ( NULL != (ptr=strrchr(headerFile,'.')) )
      strcpy(ptr,".h");
    else
      strcat(headerFile,".h");
    strcpy(file,headerFile);
    if ( NULL != (ptr=strrchr(file,'.')) )
      strcpy(ptr,".cpp");
    else
      strcat(file,".cpp"); // add cpp extension if ommitted
    if ( NULL != (ptr=strrchr(file,'\\')) )
      strcpy(fileId,&ptr[1]);
    else
      strcpy(fileId,file);
    if ( NULL != (ptr=strrchr(fileId,'.')) )
      ptr[0]='\0';
    for (int i=strlen(fileId)-1;i>=0;i--)
    {
      if ( ! isalnum(fileId[i]) )
        fileId[i]='_';
    }
    int len=ExportFile(buff,fileId,unixLinefeeds,false,!excludeMain,false,true);
    HANDLE handle=CreateFile(file,GENERIC_WRITE|GENERIC_READ,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
    if (INVALID_HANDLE_VALUE!=handle)
    {
      DWORD bytesProcessed=0;
      --len; // don't write null byte out
      if ( 0!=WriteFile(handle,buff,len,&bytesProcessed,NULL) && len==bytesProcessed )
        success=true;
    }
    CloseHandle(handle);
    len=ExportFile(buff,fileId,unixLinefeeds,true,false,true,false);
    handle=CreateFile(headerFile,GENERIC_WRITE|GENERIC_READ,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);
    if (INVALID_HANDLE_VALUE!=handle)
    {
      DWORD bytesProcessed=0;
      --len; // don't write null byte out
      if ( 0!=WriteFile(handle,buff,len,&bytesProcessed,NULL) && len==bytesProcessed )
        success=true;
    }
    CloseHandle(handle);
  }
  else // place data on clipboard output
  {
    int len=ExportFile(buff,"",false,unixLinefeeds,!excludeMain,true,true);
    HGLOBAL hGlobalMemory=GlobalAlloc(GHND,(DWORD)len+128);
    if (hGlobalMemory!=NULL)
    {
      char far * lpGlobalMemory=(char far *)GlobalLock(hGlobalMemory);
      strcpy(lpGlobalMemory,buff);
      GlobalUnlock(hGlobalMemory);
      OpenClipboard(NULL);
      EmptyClipboard();
      SetClipboardData(CF_TEXT,hGlobalMemory);
      CloseClipboard();
      success=true;
    }
  }
  delete[] buff;
  return success;
}
//---------------------------------------------------------------------------
void CFormGroup::Swap(int a,int b)
{
  // swap two objects in list
  if ( a>0 && a<m_totalObj && b>0 && b<m_totalObj && a!=b )
  {
    CBaseObject *temp=m_obj[a];
    m_obj[a]=m_obj[b];
    m_obj[b]=temp;
    m_dirty=true;
  }
}
//---------------------------------------------------------------------------
int CFormGroup::ExportFile(char *ptr,
                           const char *fileName,
                           bool unixLinefeeds,
                           bool outputHeader,
                           bool outputMain,
                           bool outputClass,
                           bool outputCode)
{
  // output form data in style suitable for compiling
  char buff[256]="";
  CObjForm *obj=dynamic_cast<CObjForm*>(m_obj[0]);
  if (NULL==obj)
    return 0;
  int left=0, top=0, width=0, height=0, i=0;
  obj->GetBounds(&left,&top,&width,&height);
  const char *formWindowName=obj->GetText(CObjForm::INDEX_NAME);
  const char *formClassName=obj->GetText(CObjForm::INDEX_CLASS);
  int pos=0;
  // wsprintf return string length created - we keep track of it to know where
  // to write nect, and total size of buffer
  if ( outputHeader ) // header file?
  {
    pos+=wsprintf(&ptr[pos],
        "\r\n#ifndef _%s_H"
        "\r\n#define _%s_H"
        "\r\n/*=============================================================================="
        "\r\n"
        "\r\n"
        "\r\n"
        "\r\n==============================================================================*/"
        "\r\n#include \"gui.h\""
        "\r\n"
        ,fileName
        ,fileName
      );
  }
  else // main (.c/.cpp) file
  {
    pos+=wsprintf(&ptr[pos],
        "/*=============================================================================="
        "\r\n"
        "\r\n"
        "\r\n"
        "\r\n==============================================================================*/"
        "\r\n#include \"gui.h\""
        "\r\n"
        "\r\nextern \"C\""
        "\r\n{"
        "\r\n#include \"ereader_hostio.h\""
        "\r\n}"
        "\r\n"
      );
    if (fileName[0])
      pos+=wsprintf(&ptr[pos],"\r\n#include \"%s.h\"",fileName);
    pos+=wsprintf(&ptr[pos],"\r\n//------------------------------------------------------------------------------");
  }
  // now output class for form itself
  if ( outputClass )
  {
    // write list of enums for each form object that needs it (via NeedsHandler())
    pos+=wsprintf(&ptr[pos],
        "\r\nenum"
        "\r\n{"
        "\r\n  %s_%s=%d,"
        ,formWindowName
        ,formWindowName
        ,m_obj[0]->GetValue(CObjForm::INDEX_ENUM)
      );
    for (i=1;i<m_totalObj;i++)
    {
      if (m_obj[i]->NeedsHandler())
        pos+=wsprintf(&ptr[pos],"\r\n  %s_%s,",formWindowName,m_obj[i]->GetName());
    }
    pos+=wsprintf(&ptr[pos],
        "\r\n};"
        "\r\n"
      );
    // now write directives text at this point
    pos+=wsprintf(&ptr[pos],"\r\n\r\n");
    i=0;
    while ( '\0' != (ptr[pos]=m_directivesText[i++]) )
      ++pos;
    pos+=wsprintf(&ptr[pos],"\r\n");
    pos+=wsprintf(&ptr[pos],
        "\r\nclass %s : public %s"
        "\r\n{"
        "\r\n"
        "\r\n  private:"
        "\r\n"
        "\r\n  public:"
        "\r\n"
        ,formWindowName
        ,formClassName
      );
    // now write object variables for any that need it
    for (i=1;i<m_totalObj;i++) // now any object variables
    {
      if (m_obj[i]->NeedsHandler())
        pos+=wsprintf(&ptr[pos],"\r\n    %s *%s;",m_obj[i]->ClassName(),m_obj[i]->GetName());
    }
    // now write out classtext info
    pos+=wsprintf(&ptr[pos],
        "\r\n"
        "\r\n    %s();"
        "\r\n"
        ,formWindowName
      );
    i=0;
    while ( '\0' != (ptr[pos]=m_classText[i++]) )
      ++pos;
    pos+=wsprintf(&ptr[pos],
        "\r\n"
        "\r\n};"
      );
  }
  // write code calls if needed
  if ( outputCode )
  {
    // output form's title and caption
    if (obj->GetValue(CObjForm::INDEX_TITLEBAR))
      wsprintf(buff,"\"%s\"",obj->GetText(CObjForm::INDEX_CAPTION));
    else
      strcpy(buff,"0");
    // form constructor
    pos+=wsprintf(&ptr[pos],
        "\r\n//------------------------------------------------------------------------------"
        "\r\n%s::%s()"
        "\r\n : CWindow( %s_%s,%d,%d,%d,%d,%s,%s,%s,%s)"
        "\r\n{"
        "\r\n  GUI_SetOrientation(%s);"
        ,formWindowName,formWindowName,formWindowName,formWindowName
        ,left,top,width,height
        ,buff
        ,( obj->GetValue(CObjForm::INDEX_SAVEUNDER)   ? "TRUE" : "FALSE" )
        ,( obj->GetValue(CObjForm::INDEX_VISIBLE)     ? "FALSE" : "TRUE" )
        ,( obj->GetValue(CObjForm::INDEX_CLOSEBUTTON) ? "TRUE" : "FALSE" )
        ,( obj->GetValue(CObjForm::INDEX_PORTRAIT_O)  ? "TRUE" : "FALSE" )
      );
    // constructor init for each object
    for (i=1;i<m_totalObj;i++)
    {
      m_obj[i]->GetBounds(&left,&top,&width,&height);
      if (m_obj[i]->NeedsHandler())
      {
        pos+=wsprintf(&ptr[pos],"\r\n  %s=new %s(%s_%s,%s);",
            m_obj[i]->GetName(),m_obj[i]->ClassName(),formWindowName,m_obj[i]->GetName(),m_obj[i]->ClassConstructorString(buff));
        pos+=wsprintf(&ptr[pos],"\r\n  AddChild(%s,%d,%d);",
            m_obj[i]->GetName(),left,top);
      }
      else
      {
        pos+=wsprintf(&ptr[pos],
            "\r\n  AddChild(new %s(1,%s),%d,%d);",
            m_obj[i]->ClassName(),m_obj[i]->ClassConstructorString(buff),left,top
          );
      }
    }
    // check m_classText for 'ConstructorInitialization' - embedded text that
    // allows us to add a call in constructor transparently
    const char *callName="ConstructorInitialization"; // long, awkward, helps prevent accidents
    if (NULL!=strstr(m_classText,callName)) // found?
      pos+=wsprintf(&ptr[pos],"\r\n  ConstructorInitialization();");
    pos+=wsprintf(&ptr[pos],"\r\n}");
  }
  // if main routine requested, write it too
  if ( outputMain )
  {
    pos+=wsprintf(&ptr[pos],
        "\r\n/*=============================================================================="
        "\r\n"
        "\r\n==============================================================================*/"
        "\r\nstatic %s *f_mainWindow;"
        "\r\n//------------------------------------------------------------------------------"
        "\r\nS32 GUI_main( MSG_TYPE type, CViewable *object, S32 data )"
        "\r\n{"
        "\r\n  switch (type)"
        "\r\n  {"
        "\r\n    case MSG_APP_START:"
        "\r\n      if (OS_is_present && hostIO_am_I_the_current_task())"
        "\r\n        HOSTIO_INLINE_BREAKPOINT();"
        "\r\n      f_mainWindow = new %s();"
        "\r\n      return 1;"
        "\r\n    case MSG_APP_STOP: // prog exit & cleanup"
        "\r\n      delete f_mainWindow;"
        "\r\n      return 1;"
        "\r\n    default:"
        "\r\n      break;"
        "\r\n  }"
        "\r\n  return 0;"
        "\r\n};"
        "\r\n//------------------------------------------------------------------------------"
        "\r\n"
        "\r\n"
        "\r\n"
        ,formWindowName,formWindowName
      );
  }
  // now finish up header
  if ( outputHeader )
    pos+=wsprintf(&ptr[pos],
        "\r\n//------------------------------------------------------------------------------"
        "\r\n#endif"
        "\r\n"
      );
  // convert file to unix style - move all text except '\n' forward, erasing them 
  if (unixLinefeeds) // remove extra char?
  {
    int pos2=0;
    for (i=0;i<=pos;i++)
    {
      if ('\n'!=ptr[i])
        ptr[pos2++]=ptr[i];
    }
    pos=pos2-1;
  }
  return pos+1; // full len, plus null byte
}
//---------------------------------------------------------------------------
































