/*===========================================================================


       main.cpp -     main vcl form handler code

                      Copyright  2001, 2003 by Pankhurst Algorithmics
                      (www.pankhurst.com, www.ebmDevMag.com)
                      All Rights Reserved


 ** LICENSE ********************************************************************
 *                                                                             *
 *  Version: MPL 1.1                                                           *
 *                                                                             *
 *  The contents of this file are subject to the Mozilla Public License        *
 *  Version 1.1 (the "License"); you may not use this file except in           *
 *  compliance with the License. You may obtain a copy of the License at       *
 *  http://www.mozilla.org/MPL/                                                *
 *                                                                             *
 *  Software distributed under the License is distributed on an "AS IS" basis, *
 *  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License   *
 *  for the specific language governing rights and limitations under the       *
 *  License.                                                                   *
 *                                                                             *
 *  The Original Code is: eBookMan Form Editor                                 *
 *                                                                             *
 *  The Initial Developer of the Original Code is                              *
 *  Pankhurst Algorithmics                                                     *
 *                                                                             *
 *  Portions created by the Initial Developer are Copyright (C) 2001-2003      *
 *  Pankhurst Algorithmics. All Rights Reserved.                               *
 *                                                                             *
 *                                                                             *
 *******************************************************************************

 
maintenance:

v1.2 (COSMETIC ONLY)
--------------------
-added commentary to code, copyright/license to each source form

v1.1
----
-fixed bug with header/class output - printf limited in output, so long text strings were being clipped
-added vertical scrollbar to header/class text dialog
-added saving of current file directory for ease of reusing program (open/save dialogs)
-added quick update button - save and export with current settings
-added status bar with display of current file
-cleared filename on new, so no resaving over previous file
-added backup of old .FFF file to .FF0 file before saving
-warn if file selected in export is not c/cpp source
-added auto-init func call - if class text defines 'ConstructorInitialization' then this call is inserted at end of constructor

v1.0
----
initial release

===========================================================================*/
#include <vcl.h>
#pragma hdrstop

#include <stdlib.h>
#include <stdio.h>

#include "objcreat.h"
#include "dlgCpp.h"
#include "about.h"
#include "formgrp.h"
#include "text_dlg.h"

#include "main.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
#pragma resource "*.dfm"
TBackPanel *BackPanel;
//---------------------------------------------------------------------------
#define FILE_BACKUP_EXTENSION ".FF0"
#define FILE_EXTENSION ".FFF"
#define FILE_EXTENSION_FILTER "*" FILE_EXTENSION
#define SOURCE_EXTENSION "*.CPP"
//---------------------------------------------------------------------------
const int c_bigSide=240, c_littleSide=200; // startup display size
const TColor c_backplaneColor=0x8FB82E;    // rgb color of backplane
const TColor c_outline=clFuchsia;          // drawing color
//---------------------------------------------------------------------------
#define FONT1 "ebm_def.fnt"
//---------------------------------------------------------------------------
__fastcall TBackPanel::TBackPanel(TComponent* Owner)
  : TForm(Owner),
    m_obj(NULL),
    m_displayDX(c_littleSide),
    m_displayDY(c_bigSide),
    m_scale(1),
    m_image(NULL),
    m_grid(4),
    m_formLeftDX(2),
    m_formTopDY(16),
    m_formOffsetX(m_formLeftDX),
    m_formOffsetY(m_formTopDY),
    m_mouseDownX(0),
    m_mouseDownY(0),
    m_mouseDragSection(DRAG_UNDEFINED),
    m_mouseDown(false),
    m_selectedIndex(-1),
    m_mouseOverIndex(-1)
{
  // form init
  m_file[0]='\0';
  m_obj=new CFormGroup(c_maxObj,"CWindow1",0,0,c_littleSide,c_bigSide);
  GRID_PROPERTIES->RowCount=0;
  // set up our ebm-style drawing font
  if (0==AddFontResource(FONT1))
    MessageBox(Handle,"Cannot add font","Error",MB_OK);
  m_selectedIndex=0;
  SyncPropertiesToObject();
  m_selectedIndex=-1;
  int width=0, height=0;
  m_image=new TImage(NULL); // get image buffer - we draw on this, and then copy to screen - allows resizing/scaling
  m_image->Width =c_bigSide;
  m_image->Height=c_bigSide;
  OpenDialog1->DefaultExt=FILE_EXTENSION_FILTER;
  OpenDialog1->Filter="Franklin Form Files (" FILE_EXTENSION_FILTER ")|" FILE_EXTENSION_FILTER ;
  // dp 1.1 - added code to save/load curr file directory
  char path[260]="", iniPath[260]="", *ptr=NULL;
  ::GetModuleFileName(NULL,iniPath,sizeof(iniPath));
  if ( NULL == (ptr=strrchr(iniPath,'.')) )
    strcat(iniPath,".INI");
  else
    strcpy(ptr,".INI");
  ::GetPrivateProfileString("PATH","LAST_FILE","",path,sizeof(path),iniPath);
  // validate/parse file to get current load directory - if invalid, don't load it
  if ( NULL != (ptr=strrchr(path,'\\')) )
    ptr[1]='\0';
  if ( 0 == (::GetFileAttributes(path)^FILE_ATTRIBUTE_DIRECTORY) ) // if valid dir (and only if) this should be zero
  {
    OpenDialog1->InitialDir=path;
    SaveDialog1->InitialDir=path;
  }
}
//---------------------------------------------------------------------------
__fastcall TBackPanel::~TBackPanel(void)
{
  // cleanup
  RemoveFontResource(FONT1);
  delete m_obj;
  delete m_image;
  // dp 1.1 - save last file for next dir info
  char path[260]="", iniPath[260]="", *ptr=NULL;
  ::GetModuleFileName(NULL,iniPath,sizeof(iniPath));
  if ( NULL == (ptr=strrchr(iniPath,'.')) )
    strcat(iniPath,".INI");
  else
    strcpy(ptr,".INI");
  if (strlen(OpenDialog1->FileName.c_str()))
    ::WritePrivateProfileString("PATH","LAST_FILE",OpenDialog1->FileName.c_str(),iniPath);
  else if (strlen(SaveDialog1->FileName.c_str()))
    ::WritePrivateProfileString("PATH","LAST_FILE",SaveDialog1->FileName.c_str(),iniPath);
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::FormCanResize(TObject *Sender, int &NewWidth,
      int &NewHeight, bool &Resize)
{
  // prevent the form from getting too small
  if (NewWidth<Label1->Left+Label1->Width+PANEL_PROPERTIES->Left+10)
    NewWidth=Label1->Left+Label1->Width+PANEL_PROPERTIES->Left+10;
  if (NewHeight<WORKSPACE_OUTLINE->Top+WORKSPACE_OUTLINE->Height+WORKSPACE_OUTLINE->Top+10)
    NewHeight=WORKSPACE_OUTLINE->Top+WORKSPACE_OUTLINE->Height+WORKSPACE_OUTLINE->Top+10;
}
//---------------------------------------------------------------------------
void TBackPanel::AdjustComponents(void)
{
  // general resizing code for every component that can be affected by a move
  int scrollbarWidth=GetSystemMetrics(SM_CXVSCROLL);
  GRID_PROPERTIES->Width =PANEL_PROPERTIES->Width-GRID_PROPERTIES->Left*2;
  GRID_PROPERTIES->Height=PANEL_PROPERTIES->Height-GRID_PROPERTIES->Left-GRID_PROPERTIES->Top;
  GRID_PROPERTIES->DefaultColWidth=(GRID_PROPERTIES->Width-scrollbarWidth-10)/2+1;
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::FormActivate(TObject *Sender)
{
  static bool firstTime=true;
  if (firstTime) // do adjustments that aren't being handled whe form initializes full size
  {
    firstTime=false;
    AdjustComponents();
    GRID_PROPERTIES->DefaultRowHeight=GRID_PROPERTIES->Canvas->TextHeight("Xy")*4/3; // and some room for bottom of letters
  }
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::FormResize(TObject *Sender)
{
  AdjustComponents();
}
//---------------------------------------------------------------------------
void TBackPanel::NameObject(int index)
{
  // name the new object uniquely, by taking class name and adding numeric
  // suffix. For that, we go through the list and find the lowest unused
  // suffix available for that object
  const char *name=m_obj->Object(index)->GetName(); // default name for object creation - no number following
  int i=0, value=0, len=strlen(name), used[c_maxObj+5];
  for (i=0;i<sizeof(used)/sizeof(used[0]);i++)
    used[i]=0;
  // now set flags for all 'used' names, leaving unused ones in array for easy scan
  for (i=0;i<m_obj->TotalObj();i++)
  {
    if ( memicmp(m_obj->Object(i)->GetName(),name,len) ) // first part not same (ie not default or not same type)
      continue;
    if ( (value=m_obj->Object(i)->DefaultName()) < 0 )
      continue; // not default
    if ( value>=0 && value<c_maxObj )
      used[value]=1;
  }
  for (i=1;i<c_maxObj;i++)
  {
    if (!used[i])
    {
      char buff[99]="";
      wsprintf(buff,"%s%d",name,i);
      m_obj->Object(index)->SetName(buff);
      m_obj->SetDirtyFlag();
      return;
    }
  }
}
//---------------------------------------------------------------------------
int TBackPanel::GetNewObject(int x,int y)
{
  // check speedbuttons - if any down, then return new default object of that type
  if (m_obj->TotalObj()>=c_maxObj)
  {
    MessageBox(Handle,"Cannot add more objects to form - limit reached","Too Many Objects",MB_OK);
    return -1;
  }
  TSpeedButton *obj=NULL;
  for (int i=BackPanel->ComponentCount-1;i>=0;i--)
  {
    obj=dynamic_cast<TSpeedButton*>(BackPanel->Components[i]);
    if ( NULL!=obj && obj->Down ) // our selection?
    {
      obj->Down=false;
      TPanel *parent=WORKSPACE;
      TControl *newObj=NULL;
      bool success=true;
      // create a new entry
      int index=m_obj->TotalObj();
      switch (obj->Tag)
      {
        case 0: // button
          m_obj->Add(new CObjButton(),x,y,55,20);
          break;
        case 1: // check
          m_obj->Add(new CObjCheckBox(),x,y,55,15);
          break;
        case 2: // textbox
          m_obj->Add(new CObjTextEdit(),x,y,40,20);
          break;
        case 4: // scrollbar
          m_obj->Add(new CObjScrollBar(),x,y,50,14);
          break;
        case 5: // text
          m_obj->Add(new CObjLabel(),x,y,60,18);
          break;
        case 6: // Table
          m_obj->Add(new CObjTable(),x,y,50,25);
          break;
        case 7: // RepeatButton
          m_obj->Add(new CObjRepeatButton(),x,y,100,20);
          break;
        case 8: // PushButton
          m_obj->Add(new CObjPushButton(),x,y,90,20);
          break;
        case 9: // ProgressBar
          m_obj->Add(new CObjProgressBar(),x,y,50,14);
          break;
        case 10: // Rect
          m_obj->Add(new CObjRect(),x,y,50,25);
          break;
        case 11: // Icon
          m_obj->Add(new CObjIcon(),x,y,50,25);
          break;
        case 13: // TimeEdit
          m_obj->Add(new CObjTimeEdit(),x,y,50,25);
          break;
        case 14: // DateEdit
          m_obj->Add(new CObjDateEdit(),x,y,50,25);
          break;
        case 15: // radiobutton
          m_obj->Add(new CObjRadioButton(),x,y,55,15);
          break;
        default:
          success=false;
          break;
      }
      if (success)
      {
        NameObject(index);
        DrawDisplay();
        return m_obj->TotalObj()-1;
      }
    }
  }
  return -1;
}
//---------------------------------------------------------------------------
void TBackPanel::SyncPropertiesToObject(void)
{
  // make display on left match list on right
  if (m_selectedIndex<0)
    return;
  int index=m_selectedIndex;
  int rows=m_obj->Object(index)->TotalProperty(), i=0;
  GRID_PROPERTIES->RowCount=rows;
  for (i=0;i<rows;i++)
  {
    GRID_PROPERTIES->Cells[0][i]=m_obj->Object(index)->PropertyName(i);
    GRID_PROPERTIES->Cells[1][i]=m_obj->Object(index)->GetText(i);
  }
  GRID_PROPERTIES->Tag=index;
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::DISPLAY_CANVASPaint(TObject *Sender)
{
  // handle repainting
  DrawDisplay();
}
//---------------------------------------------------------------------------
void TBackPanel::AdjustFormPosition(void)
{
  // adjust form offset values
  CObjForm *obj=dynamic_cast<CObjForm*>(m_obj->Object(0));
  int left=0, top=0, width=0, height=0, i=0;
  obj->GetBounds(&left,&top,&width,&height);
  m_formLeftDX=2;
  m_formOffsetX=left+m_formLeftDX;
  if (obj->GetValue(CObjForm::INDEX_TITLEBAR))
    m_formTopDY=16;
  else
    m_formTopDY=2;
  m_formOffsetY=top+m_formTopDY;
}
//---------------------------------------------------------------------------
void TBackPanel::DrawDisplay(void)
{
  // draw form & all objects - double-buffered to avoid flicker
  AdjustFormPosition();
  TRect srcRect(0,0,m_displayDX,m_displayDY), dstRect(0,0,srcRect.right*m_scale,srcRect.bottom*m_scale);
  m_image->Canvas->Brush->Color=c_backplaneColor;
  m_image->Canvas->FillRect(srcRect);
  int left=0, top=0, width=0, height=0, i=0;
  for (int i=0;i<m_obj->TotalObj();i++)
  {
    if (i) // draw all form children objects
      m_obj->Object(i)->Draw(m_image->Canvas,m_formOffsetX,m_formOffsetY);
    else // i==0 - form object
      m_obj->Object(i)->Draw(m_image->Canvas,0,0);
    if (i==m_selectedIndex) // draw outline rectangle?
    {
      m_image->Canvas->Pen->Color=c_outline;
      m_image->Canvas->Pen->Style=psDot;
      m_image->Canvas->Brush->Style=bsClear;
      int left=0, top=0, width=0, height=0;
      m_obj->Object(i)->GetBounds(&left,&top,&width,&height);
      if (i) // all but main form gets offsets added on to draw outline
      {
        left+=m_formOffsetX;
        top +=m_formOffsetY;
      }
      m_image->Canvas->MoveTo(left-2,      top-2       );
      m_image->Canvas->LineTo(left-2,      top+height+1);
      m_image->Canvas->LineTo(left+width+1,top+height+1);
      m_image->Canvas->LineTo(left+width+1,top-2       );
      m_image->Canvas->LineTo(left-2,      top-2       );
    }
  }
  // redraw form frame to cover/clip other children drawing
  CObjForm *obj=dynamic_cast<CObjForm*>(m_obj->Object(0));
  obj->NonClientDraw(m_image->Canvas,0,0);
  // draw finished form to screen (and scale as needed)
  DISPLAY_CANVAS->Canvas->CopyRect(dstRect,m_image->Canvas,srcRect);
}
//---------------------------------------------------------------------------
int TBackPanel::GetSelectedObject(int x,int y)
{
  // return index of found object at position (x,y), or -1 if not found
  int left=0, top=0, width=0, height=0, i=0;
  for (i=m_obj->TotalObj()-1;i>=0;i--)
  {
    m_obj->Object(i)->GetBounds(&left,&top,&width,&height);
    if (i)
    {
      left+=m_formOffsetX;
      top +=m_formOffsetY;
    }
    if ( x>=left && x<=width+left && y>=top && y<=height+top )
      return i;
  }
  return -1;
}
//---------------------------------------------------------------------------
void TBackPanel::AdjustDisplaySize(int newScale)
{
  // resize display
  int dx=Width-ClientWidth;
  int dy=Height-ClientHeight;
  int w=dx+m_displayDX*newScale+WORKSPACE_OUTLINE->Left*2+PANEL_PROPERTIES->Width;
  if (w>Screen->Width)
    PANEL_PROPERTIES->Width=Screen->Width-dx-m_displayDX*newScale-WORKSPACE_OUTLINE->Left*2;
  if ( WindowState!=wsMaximized && Left+w>Screen->Width )
    Left=Screen->Width-w;
  WORKSPACE_BACK->Width=m_displayDX*newScale;
  if (WindowState!=wsMaximized)
    Width=w;
  int h=dy+m_displayDY*newScale+WORKSPACE_OUTLINE->Top+WORKSPACE_OUTLINE->Left;
  if (h>Screen->Height)
    h=Screen->Height;
  if (WindowState!=wsMaximized)
    Height=h;
  if ( WindowState!=wsMaximized && Top+h>Screen->Height )
    Top=Screen->Height-h;
  WORKSPACE_BACK->Width=m_displayDX*newScale+WORKSPACE_OUTLINE->Left*2+dx/2;
  WORKSPACE->Width =m_displayDX*newScale;
  WORKSPACE->Height=m_displayDY*newScale;
  if (ClientHeight<WORKSPACE->Top+WORKSPACE->Height+10)
    ClientHeight=WORKSPACE->Top+WORKSPACE->Height+10;
  WORKSPACE_OUTLINE->Width =WORKSPACE->Width+WORKSPACE_OUTLINE->Pen->Width*2;
  WORKSPACE_OUTLINE->Height=WORKSPACE->Height+WORKSPACE_OUTLINE->Pen->Width*2;
  AdjustComponents();
  m_scale=newScale;
  // dp 1.1 - adjust statusbar
  StatusBar1->Top=ClientHeight-StatusBar1->Height;
}
/*===========================================================================

       Menu Option Handlers

===========================================================================*/
void __fastcall TBackPanel::QuickUpdate1Click(TObject *Sender)
{
  // dp 1.1  - added to perform save/export (at current settings) together
  SaveDialog1->DefaultExt=FILE_EXTENSION_FILTER;
  SaveDialog1->Filter="Franklin Form Files (" FILE_EXTENSION_FILTER ")|" FILE_EXTENSION_FILTER ;
  char *caption="Quick Update";
  if ( '\0'==m_file[0] )
    SaveAs1Click(Sender);
  else if (!SaveWithBackup(m_file))
    MessageBox(Handle,"ERROR Saving!",caption,MB_OK);
  // check if we have a export file selected, or are we sending to clipboard
  if (0==strlen(dlgCodeSettings->Label1->Caption.c_str()))
  {
    if (ID_YES!=::MessageBox(Handle,"You are currently set to export to the clipboard - "
                                   "is that what you want?",caption,MB_YESNO))
    {
      ExportCode1Click(Sender);
      return;
    }
  }
  // now export at current settings
  if (!dlgCodeSettings->Execute(m_obj,true))
    MessageBox(Handle,"ERROR Exporting!",caption,MB_OK);
  else
    MessageBox(Handle,"Saved and exported successfully",caption,MB_OK);
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::NormalSize1Click(TObject *Sender)
{
  TComponent *obj=dynamic_cast<TComponent*>(Sender);
  if (NULL!=obj)
    AdjustDisplaySize(obj->Tag);
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::Grid1Click(TObject *Sender)
{
  // display input box and get new grid size
  AnsiString NewString = m_grid;
  if (InputQuery("Set Grid Size", "Grid size in pixels (1-20):", NewString))
  {
    int value=NewString.ToInt();
    if (value<1||value>20)
      Application->MessageBox("Invalid Value","Grid value out of range",MB_OK);
    else
      m_grid=value;
    UseGrid1->Checked=true;
  }
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::UseGrid1Click(TObject *Sender)
{
  // toggle grid setting
  if (UseGrid1->Checked)
    UseGrid1->Checked=false;
  else
    UseGrid1->Checked=true;
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::About1Click(TObject *Sender)
{
  AboutBox->ShowModal();
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::Exit1Click(TObject *Sender)
{
  Close();
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::MovetoBack1Click(TObject *Sender)
{
  // promote selected item to top of list ('on top') of view)
  int index=m_selectedIndex;
  const char *caption="Error";
  if (index<0)
    MessageBox(Handle,"Nothing selected",caption,MB_OK);
  else if (0==index)
    MessageBox(Handle,"Can't move main form",caption,MB_OK);
  else if (ID_YES==MessageBox(Handle,"Move Highlighted Item?","Move to Back",MB_YESNO))
  {
    for (int i=index-1;i>0;i--)
      m_obj->Swap(i,i+1);
    m_selectedIndex=0;
    SyncPropertiesToObject();
    DrawDisplay();
  }
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::SendtoFront1Click(TObject *Sender)
{
  // promote selected item to top of list ('on top') of view)
  int index=m_selectedIndex;
  const char *caption="Error";
  if (index<0)
    MessageBox(Handle,"Nothing selected",caption,MB_OK);
  else if (0==index)
    MessageBox(Handle,"Can't move main form",caption,MB_OK);
  else if (ID_YES==MessageBox(Handle,"Move Highlighted Item?","Move to Front",MB_YESNO))
  {
    for (int i=index;i<m_obj->TotalObj()-1;i++)
      m_obj->Swap(i,i+1);
    m_selectedIndex=0;
    SyncPropertiesToObject();
    DrawDisplay();
  }
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::menuHelpClick(TObject *Sender)
{
  // display help file
  char dir[256]="", *ptr=NULL;
  ::GetModuleFileName(NULL,dir,sizeof(dir));
  if ( NULL != (ptr=strrchr(dir,'.')) )
    ptr[1]='\0';
  strcat(dir,"HLP");
  ::WinHelp(Handle,dir,HELP_CONTENTS,0);
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::ExportCode1Click(TObject *Sender)
{
  if (dlgCodeSettings->Execute(m_obj,false))
    MessageBox(Handle,"Done","Export C++ Source",MB_OK);
}
//---------------------------------------------------------------------------
bool TBackPanel::NeedToSave(void)
{
  // test and ask if changes made
  if ( NULL==m_obj || !m_obj->Dirty() || m_obj->TotalObj()<=1 || ID_YES==MessageBox(Handle,"Current file not saved! Continue anyways?","Unsaved File",MB_YESNO) )
    return false;
  return true;
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::FormCloseQuery(TObject *Sender, bool &CanClose)
{
  CanClose=!NeedToSave();
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::NewForm1Click(TObject *Sender)
{
  // replace form
  if (!NeedToSave())
  {
    delete m_obj;
    m_obj=new CFormGroup(c_maxObj,"CForm1",0,0,c_littleSide,c_bigSide);
    m_selectedIndex=0;
    SyncPropertiesToObject();
    DrawDisplay();
    m_file[0]='\0'; // erase old filename
    dlgCodeSettings->Label1->Caption=""; // clear export name so we don't accidently write to wrong file
    DrawStatus();
  }
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::Directives1Click(TObject *Sender)
{
  // add directives text
  if (dlgText->Execute("Class Text",&m_obj->m_classText))
    m_obj->SetDirtyFlag();
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::Source1Click(TObject *Sender)
{
  // add header text
  if (dlgText->Execute("Header Text",&m_obj->m_directivesText))
    m_obj->SetDirtyFlag();
}
/*===========================================================================


===========================================================================*/
void TBackPanel::HandleGridEntry(void)
{
  // edit data to grid
  static bool locked=false;
  if (locked) // avoid reentrancy as we change values
    return;
  int index=GRID_PROPERTIES->Tag, i=0, row=GRID_PROPERTIES->Row, col=GRID_PROPERTIES->Col;
  if ( index<0 )
    return;
  char buff[256]="";
  strcpy(buff,GRID_PROPERTIES->Cells[col][row].c_str());
  if (m_obj->Object(index)->IndexName()==GRID_PROPERTIES->Row) // trying to change name?
  {
    for (i=0;i<m_obj->TotalObj();i++) // not duplicated elsewhere is it?
    {
      if ( i!=index && 0==stricmp(buff,m_obj->Object(i)->GetName()) )
      {
        MessageBox(Handle,"Duplicate name on form - not allowed","Entry Error",MB_OK);
        locked=true;
        GRID_PROPERTIES->Cells[col][row]=m_obj->Object(index)->GetText(row);
        locked=false;
        return;
      }
    }
  }
  bool success=m_obj->Object(index)->TestAndSet(row,buff);
  m_obj->SetDirtyFlag();
  locked=true;
  GRID_PROPERTIES->Cells[col][row]=m_obj->Object(index)->GetText(row);
  locked=false;
  if (!success)
    return;
  if (0==index) // base form? then adjust all other objects before draw
  {
    int left=0, top=0, width=0, height=0, i=0;
    m_obj->Object(0)->GetBounds(&left,&top,&width,&height);
    CheckOrientation();
  }
  DrawDisplay();
}
//---------------------------------------------------------------------------
void TBackPanel::CheckOrientation(void)
{
  // if orientation adjusted, we need to change display as well...
  CObjForm *obj=dynamic_cast<CObjForm*>(m_obj->Object(0));
  bool portrait=obj->GetValue(CObjForm::INDEX_PORTRAIT_O);
  if ( m_displayDX==c_bigSide && portrait || m_displayDX!=c_bigSide && !portrait )
  {
    if (portrait)
    {
      m_displayDX=c_littleSide;
      m_displayDY=c_bigSide;
    }
    else // change it
    {
      m_displayDX=c_bigSide;
      m_displayDY=c_littleSide;
    }
    AdjustDisplaySize(m_scale);
  }
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::GRID_PROPERTIESSelectCell(TObject *Sender,
      int ACol, int ARow, bool &CanSelect)
{
  // handle cell selection
  if (0!=ACol)
    GRID_PROPERTIES->Options=GRID_PROPERTIES->Options<<goEditing;
  else
    GRID_PROPERTIES->Options=GRID_PROPERTIES->Options>>goEditing;
  if ( 1==GRID_PROPERTIES->Col && ( ACol!=GRID_PROPERTIES->Col || ARow!=GRID_PROPERTIES->Row ) )
    HandleGridEntry();
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::GRID_PROPERTIESKeyPress(TObject *Sender,
      char &Key)
{
  // manage keypresses
  if (0==GRID_PROPERTIES->Col) // can't enter text in first col
  {
    Key=0;
    return;
  }
  if (VK_RETURN==Key) // entered?
  {
    HandleGridEntry();
  }
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::LockForm1Click(TObject *Sender)
{
  // lock mode toggle
  if (LockForm1->Checked)
  {
    LockForm1->Checked=false;
    LockForm1->Caption="Lock Form";
  }
  else
  {
    LockForm1->Checked=true;
    LockForm1->Caption="Unlock Form";
  }
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::DeleteItem1Click(TObject *Sender)
{
  // handle deletion request
  int index=m_selectedIndex;
  if (index<0)
    MessageBox(Handle,"Nothing selected","Delete Error",MB_OK);
  else if (0==index)
    MessageBox(Handle,"Can't delete main form","Delete Error",MB_OK);
  else if (ID_YES==MessageBox(Handle,"Delete Highlighted Item?","Delete",MB_YESNO))
  {
    m_obj->Delete(index);
    m_selectedIndex=0;
    SyncPropertiesToObject();
    DrawDisplay();
  }
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::GRID_PROPERTIESDblClick(TObject *Sender)
{
  // double-click handler 
  int index=GRID_PROPERTIES->Tag, i=0, row=GRID_PROPERTIES->Row, col=GRID_PROPERTIES->Col;
  if ( 1!=col || index<0 )
    return;
  if (m_obj->Object(index)->Editor(row))
  {
    GRID_PROPERTIES->Cells[1][row]=m_obj->Object(index)->GetText(row);
    DISPLAY_CANVAS->Invalidate();
  }
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::SaveAs1Click(TObject *Sender)
{
  // save as handler
  SaveDialog1->DefaultExt=FILE_EXTENSION_FILTER;
  SaveDialog1->Filter="Franklin Form Files (" FILE_EXTENSION_FILTER ")|" FILE_EXTENSION_FILTER ;
  if (!SaveDialog1->Execute())
    return;
  if (SaveWithBackup(SaveDialog1->FileName.c_str()))
  {
    MessageBox(Handle,"Saved","Save File",MB_OK);
    strcpy(m_file,SaveDialog1->FileName.c_str());
  }
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::Save1Click(TObject *Sender)
{
  // save handler
  SaveDialog1->DefaultExt=FILE_EXTENSION_FILTER;
  SaveDialog1->Filter="Franklin Form Files (" FILE_EXTENSION_FILTER ")|" FILE_EXTENSION_FILTER ;
  if ( '\0'==m_file[0] )
    SaveAs1Click(Sender);
  else if (SaveWithBackup(m_file))
    MessageBox(Handle,"Saved","Save File",MB_OK);
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::LoadForm1Click(TObject *Sender)
{
  // loading
  if (NeedToSave())
    return;
  if (!OpenDialog1->Execute())
    return;
  if (!m_obj->Load(OpenDialog1->FileName.c_str()))
    return; // errors
  strcpy(m_file,OpenDialog1->FileName.c_str());
  m_selectedIndex=0;
  int left=0, top=0, width=0, height=0, i=0;
  m_obj->Object(0)->GetBounds(&left,&top,&width,&height);
  CheckOrientation();
  SyncPropertiesToObject();
  DrawDisplay();
  // dp 1.1 - added current file to new status bar...
  dlgCodeSettings->Label1->Caption=""; // clear export name so we don't accidently write to wrong file
  DrawStatus();
}
/*===========================================================================

       mouse handlers

===========================================================================*/
void TBackPanel::ObjToScreenPos(int *xPos,int *yPos,int index)
{
  // convert object position to screen (panel) position
  int left=0, top=0, width=0, height=0;
  m_obj->Object(0)->GetBounds(xPos,yPos,&width,&height);
  if (index) // for non-form obj calc offset
  {
    m_obj->Object(index)->GetBounds(&left,&top,&width,&height);
    *xPos+=left+m_formLeftDX;
    *yPos+=top+m_formTopDY;
  }
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::WORKSPACEMouseUp(TObject *Sender,TMouseButton Button, TShiftState Shift, int X, int Y)
{
  m_mouseDown=false;
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::WORKSPACEMouseDown(TObject *Sender,TMouseButton Button, TShiftState Shift, int X, int Y)
{
  // take mouse down and decide if we are working on something - drop, move,
  // or drag start
  int index=-1, xPos=X/m_scale, yPos=Y/m_scale, left=0, top=0, width=0, height=0, grid=( UseGrid1->Checked ? m_grid : 1 );
  if (m_mouseOverIndex<=0)
  {
    if (0<=(index=GetNewObject( xPos-xPos%grid-m_formOffsetX, yPos-yPos%grid-m_formOffsetY ))) // dropping an object on the form?
      m_selectedIndex=index;
    else
    {
      m_obj->Object(0)->GetBounds(&left,&top,&width,&height);
      xPos=X/m_scale;
      yPos=Y/m_scale;
      if ( xPos>=left && xPos<=left+width && yPos>=top && yPos<=top+height )
        m_selectedIndex=0;
      else
        m_selectedIndex=-1;
    }
  }
  else
    m_selectedIndex=m_mouseOverIndex;
  m_mouseDown=true;
  if (m_selectedIndex>=0) // get future offsets
  {
    ObjToScreenPos(&left,&top,m_selectedIndex);
    m_mouseDownX=xPos;
    m_mouseDownY=yPos;
    m_obj->Object(m_selectedIndex)->GetBounds(&left,&top,&width,&height);
    m_selRect.left  =left;
    m_selRect.top   =top;
    m_selRect.right =width+left;
    m_selRect.bottom=height+top;
  }
  DrawDisplay();
  SyncPropertiesToObject();
}
//---------------------------------------------------------------------------
void __fastcall TBackPanel::WORKSPACEMouseMove(TObject *Sender,TShiftState Shift, int X, int Y)
{
  // take appropriate action if moving mouse
  int xPos=X/m_scale, yPos=Y/m_scale;
  // if not down, make sure cursors are set appropriately
  if (!m_mouseDown)
  {
    if ( X<0 || X>WORKSPACE->Width || Y<0 || Y>WORKSPACE->Height )
      return;
    int left=0, top=0, width=0, height=0, i=0, offset=5;
    xPos=xPos-m_formOffsetX;
    yPos=yPos-m_formOffsetY;
    m_mouseOverIndex=-1;
    m_mouseDragSection=DRAG_UNDEFINED;
    TCursor cursor=crDefault;
    for (i=0;i<m_obj->TotalObj();i++)
    {
      if (0==i)
        xPos=X/m_scale, yPos=Y/m_scale;
      m_obj->Object(i)->GetBounds(&left,&top,&width,&height);
      ObjToScreenPos(&left,&top,i);
      if ( xPos<=left-offset || xPos>=left+width+offset || yPos<=top-offset || yPos>=top+height+offset )
        continue;
      m_mouseOverIndex=i;
      if ( xPos>=left+offset && xPos<=left+width-offset && yPos>=top+offset && yPos<=top+height-offset ) // inside?
      {
        cursor=crSizeAll;
        m_mouseDragSection=DRAG_CENTER;
      }
      else if ( left-offset<xPos && xPos<left+offset ) // over left edge?
      {
        cursor=crSizeWE;
        m_mouseDragSection=DRAG_LEFT;
      }
      else if ( left+width-offset<xPos && xPos<left+width+offset ) // over right edge?
      {
        cursor=crSizeWE;
        m_mouseDragSection=DRAG_RIGHT;
      }
      else if ( top-offset<yPos && yPos<top+offset ) // over top edge?
      {
        cursor=crSizeNS;
        m_mouseDragSection=DRAG_TOP;
      }
      else // remaining - over bottom edge?
      {
        cursor=crSizeNS;
        m_mouseDragSection=DRAG_BOTTOM;
      }
    }
    if ( 0==m_mouseOverIndex && LockForm1->Checked ) // can't move, then don't allow cursor change, etc
    {
      m_mouseDragSection=DRAG_UNDEFINED;
      m_mouseOverIndex=-1;
    }
    if ( m_mouseOverIndex<0 )
      cursor=crDefault;
    WORKSPACE->Cursor=cursor;
    Application->ProcessMessages();
    return;
  }
  // mouse is down - drag appropriate section of object
  int index=m_selectedIndex, gridX=1, gridY=1, left=0, top=0, width=0, height=0, newDX=0, newDY=0;
  if (index<0)
    return;
  if ( 0==index && LockForm1->Checked )
    return;
  int grid=( UseGrid1->Checked ? m_grid : 1 );
  m_obj->Object(index)->GetBounds(&left,&top,&width,&height);
  m_obj->SetDirtyFlag();
  switch (m_mouseDragSection)
  {
    case DRAG_LEFT:
      left =m_selRect.left-m_mouseDownX+xPos;
      width=m_selRect.right-left;
      left -=left%grid;
      width-=width%grid;
      m_obj->Object(index)->SetBounds(left,top,width,height);
      break;
    case DRAG_TOP:
      top   =m_selRect.top-m_mouseDownY+yPos;
      height=m_selRect.bottom-top;
      top   -=top%grid;
      height-=height%grid;
      m_obj->Object(index)->SetBounds(left,top,width,height);
      break;
    case DRAG_RIGHT:
      width=m_selRect.right-m_selRect.left-m_mouseDownX+xPos;
      width-=width%grid;
      m_obj->Object(index)->SetBounds(left,top,width,height);
      break;
    case DRAG_BOTTOM:
      height=m_selRect.bottom-m_selRect.top-m_mouseDownY+yPos;
      height-=height%grid;
      m_obj->Object(index)->SetBounds(left,top,width,height);
      break;
    default:
    case DRAG_CENTER: // maintain invariant: distance to top left corner from cursor
      left=m_selRect.left-m_mouseDownX+xPos;
      top =m_selRect.top -m_mouseDownY+yPos;
      left-=left%grid;
      top -=top%grid;
      m_obj->Object(index)->SetBounds(left,top,width,height);
      break;
  }
  if (0==index) // base form? then adjust all other objects before draw
  {
    int left=0, top=0, width=0, height=0, i=0;
    m_obj->Object(0)->GetBounds(&left,&top,&width,&height);
  }
  DrawDisplay();
  SyncPropertiesToObject();
  return;
}
/*===========================================================================


===========================================================================*/
void TBackPanel::DrawStatus(void)
{
  // dp 1.1 - new function to display status info
  StatusBar1->SimpleText=m_file;
}
//---------------------------------------------------------------------------
bool TBackPanel::SaveWithBackup(const char *filename)
{
  // dp 1.1 - before save, renames old (if any) to backup files
  char fileWhole[260]="", filePart[260]="", *ptr=NULL;
  // get whole filename for deletion of second generation
  strcpy(fileWhole,filename);
  // now replace extension
  if ( NULL == (ptr=strrchr(fileWhole,'.')) )
    ptr=&fileWhole[strlen(fileWhole)];
  strcpy(ptr,FILE_BACKUP_EXTENSION); // replace extension
  // now take off just filename to make name needed for rename
  if ( NULL == (ptr=strrchr(fileWhole,'\\')) )
    strcpy(filePart,fileWhole);
  else
    strcpy(filePart,&ptr[1]);
  // take just filename
  ::DeleteFile(fileWhole);
  ::RenameFile(filename,filePart);
  return m_obj->Save(filename);
}
//---------------------------------------------------------------------------


