/*===========================================================================


       objcreat.cpp - windows imitations of ebm object classes  

                      Copyright  2001, 2003 by Pankhurst Algorithmics
                      (www.pankhurst.com, www.ebmDevMag.com)
                      All Rights Reserved

 ** LICENSE ********************************************************************
 *                                                                             *
 *  Version: MPL 1.1                                                           *
 *                                                                             *
 *  The contents of this file are subject to the Mozilla Public License        *
 *  Version 1.1 (the "License"); you may not use this file except in           *
 *  compliance with the License. You may obtain a copy of the License at       *
 *  http://www.mozilla.org/MPL/                                                *
 *                                                                             *
 *  Software distributed under the License is distributed on an "AS IS" basis, *
 *  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License   *
 *  for the specific language governing rights and limitations under the       *
 *  License.                                                                   *
 *                                                                             *
 *  The Original Code is: eBookMan Form Editor                                 *
 *                                                                             *
 *  The Initial Developer of the Original Code is                              *
 *  Pankhurst Algorithmics                                                     *
 *                                                                             *
 *  Portions created by the Initial Developer are Copyright (C) 2001-2003      *
 *  Pankhurst Algorithmics. All Rights Reserved.                               *
 *                                                                             *
 *                                                                             *
 *******************************************************************************


===========================================================================*/
#include <vcl.h>
#pragma hdrstop

#include "objcreat.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
const char*c_defaultFontName="ebm Default";
const int  c_defaultFontSize=8;
//---------------------------------------------------------------------------
// approximate RGB palette coloring for 16 ebm colors
const TColor long c_color[16] = // approx color values - evenly distributed here...
  {  (TColor)PALETTERGB(  0,  0,  0),  (TColor)PALETTERGB( 17, 17, 17),  (TColor)PALETTERGB( 34, 34, 34),
     (TColor)PALETTERGB( 51, 51, 51),  (TColor)PALETTERGB( 68, 68, 68),  (TColor)PALETTERGB( 85, 85, 85),
     (TColor)PALETTERGB(102,102,102),  (TColor)PALETTERGB(119,119,119),  (TColor)PALETTERGB(136,136,136),
     (TColor)PALETTERGB(153,153,153),  (TColor)PALETTERGB(170,170,170),  (TColor)PALETTERGB(187,187,187),
     (TColor)PALETTERGB(204,204,204),  (TColor)PALETTERGB(221,221,221),  (TColor)PALETTERGB(238,238,238),
     (TColor)PALETTERGB(255,255,255) };
/*===========================================================================

       CBaseObject virtual base class

===========================================================================*/
char CBaseObject::s_tempBuff[99]="";
BYTE CBaseObject::s_idsInUse[c_maxId];
//------------------------------------------------------------------------------
CBaseObject::CBaseObject(const char *className,bool needsHandler,e_objType objType,
                         const s_property *propTable,int propTableLen,
                         int indexLeft,int indexTop,int indexWidth,
                         int indexHeight,int indexName,int indexCaption,int indexClass )
 : c_property(propTable),
   c_totalProperty(propTableLen),
   m_indexLeft(indexLeft),
   m_indexTop(indexTop),
   m_indexWidth(indexWidth),
   m_indexHeight(indexHeight),
   m_indexName(indexName),
   m_indexCaption(indexCaption),
   m_indexClass(indexClass),
   m_needsHandler(needsHandler),
   m_objType(objType)
{
  // init - create property table
  m_value=new long[c_totalProperty];
  // each class is passed a property table for each child object - verify it is
  // correctly set up
  for (int i=0;i<(signed)c_totalProperty;i++)
  {
    m_value[i]=0;
    if (c_property[i].index!=i)
    {
      char buff[256]="";
      wsprintf(buff,"Error in setup - property \"%s\" %d not valid",
                    c_property[i].name,
                    c_property[i].index);
      Application->MessageBox(buff,"Table Error",MB_OK);
    }
  }
  Set(m_indexName,&className[1]); // name is same as class name, but without the initial 'C'
  if (m_indexCaption>=0)
    Set(m_indexCaption,&className[1]);
  Set(m_indexClass,className);
}
//------------------------------------------------------------------------------
CBaseObject::~CBaseObject(void)
{
  delete[] m_value;
}
//------------------------------------------------------------------------------
const char *CBaseObject::GetName(void)
{
  // get object name
  return GetText(m_indexName);
}
//------------------------------------------------------------------------------
void CBaseObject::SetName(const char*name)
{
  // set object name
  Set(m_indexName,name);
}
//------------------------------------------------------------------------------
void CBaseObject::SetBounds(long left,long top,long width,long height)
{
  // not every object uses all four bounds, but they can be set
  m_value[m_indexLeft]  =max( c_property[m_indexLeft].param1,   min( left,   c_property[m_indexLeft].param2   ));
  m_value[m_indexTop]   =max( c_property[m_indexTop].param1,    min( top,    c_property[m_indexTop].param2    ));
  m_value[m_indexWidth] =max( c_property[m_indexWidth].param1,  min( width,  c_property[m_indexWidth].param2  ));
  m_value[m_indexHeight]=max( c_property[m_indexHeight].param1, min( height, c_property[m_indexHeight].param2 ));
}
//------------------------------------------------------------------------------
void CBaseObject::GetBounds(int *left,int *top,int *width,int *height)
{
  // get bounding rectangle
  *left  =m_value[m_indexLeft];
  *top   =m_value[m_indexTop];
  *width =m_value[m_indexWidth];
  *height=m_value[m_indexHeight];
}
//------------------------------------------------------------------------------
int CBaseObject::DefaultName(void)
{
  // default name is object name, minus initial 'C', and with one or more digits
  // trailing and nothing else - otherwise, it has been customized and should
  // be treated special by constructors - if default, return value of digits
  // if not, return -1
  const char *name=GetText(m_indexName), *temp=GetText(m_indexClass), *defName=&temp[1];
  int len=strlen(defName);
  if ( memicmp(name,defName,len) ) // first part not same?
    return -1;
  const char *ptr=&name[len];
  while (isdigit(*ptr))
    ptr++;
  if ('\0'!=*ptr) // not all digits - sure sign of modified name
    return -1;
  return atoi(&name[len]);
}
//------------------------------------------------------------------------------
void CBaseObject::CleanUpTableValues(void)
{
  // delete pointer values, and generally prepare table to be deleted
  // (if dynamically allocated by child)
  for (int i=TotalProperty()-1;i>=0;i++)
  {
    if ( TYPE_CHARPTR,0,0==c_property[i].type && m_value[i] )
    {
      char *ptr=(char*)m_value[i];
      delete[] ptr;
      m_value[i]=0;
    }
  }
}
//------------------------------------------------------------------------------
void CBaseObject::Set(unsigned int propertyIndex,long value)
{
  // set property properly
  if (!ValidPropertyIndex(propertyIndex))
    return;
  if ( TYPE_CHARPTR!=c_property[propertyIndex].type ) // s/b string?
    m_value[propertyIndex]=value;
  else
  {
    char buff[39]="";
    wsprintf(buff,"%d",value);
    Set(propertyIndex,buff);
    return;
  }
}
//------------------------------------------------------------------------------
bool CBaseObject::TestAndSet(unsigned int propertyIndex,const unsigned char*ptr)
{
  // given string, check and set string - pops up dialog if invalid
  int i=propertyIndex;
  if (ValidPropertyIndex(i))
  {
    long value=atol(ptr);
    switch (c_property[i].type)
    {
      case TYPE_NUMBER:
        if ( value<c_property[i].param1 || c_property[i].param2<value )
        {
          MessageBox(NULL,"Invalid Value!","Entry Error",MB_OK);
          return false;
        }
        Set(propertyIndex,value);
        return true;
      case TYPE_BOOLEAN:
        if (!isdigit(ptr[0])) // try t, T, f, F
        {
          char *test[]={"f","false","t","true"};
          for (int i=0;i<sizeof(test)/sizeof(test[0]);i++)
          {
            if (0==stricmp(ptr,test[i]))
            {
              Set(propertyIndex,i/2); // converts trues->1, falses->0
              return true;
            }
          }
          MessageBox(NULL,"Invalid Value!","Entry Error",MB_OK);
          return false;
        }
        Set(propertyIndex,(value?1:0));
        return true;
      case TYPE_CHARPTR:
        for (int i=strlen(ptr)-1;i>=0;i--)
        {
          if ( ' '<=ptr[i] && ptr[i]<=0x7F ) // minimum amount of validity for char!
            continue;
          MessageBox(NULL,"Invalid characters in text!","Entry Error",MB_OK);
          return false;
        }
        if (1&c_property[i].param1) // names - only alphanum char, first char alpha, and only _ embedded
        {
          if ( ! ( 'a'<=*ptr && *ptr<='z' || 'A'<=*ptr && *ptr<='Z' ) )
          {
            MessageBox(NULL,"First character must be alphabetic!","Entry Error",MB_OK);
            return false;
          }
          for (int i=strlen(ptr)-1;i>=1;i--)
          {
            if ( 'a'<=ptr[i] && ptr[i]<='z'
              || 'A'<=ptr[i] && ptr[i]<='Z'
              || '0'<=ptr[i] && ptr[i]<='9'
              || '_'==ptr[i] )
                continue;
            MessageBox(NULL,"Only alphanumeric characters allowed here!","Entry Error",MB_OK);
            return false;
          }
        }
        if (2&c_property[i].param1) // non-zero length
        {
          if ('\0'==ptr[0])
          {
            MessageBox(NULL,"There must be an entry here!","Entry Error",MB_OK);
            return false; // must be some text here...
          }
        }
        Set(propertyIndex,(char*)ptr);
        return true;
      default:
        break;
    }
  }
  return false;
}
//------------------------------------------------------------------------------
bool CBaseObject::Editor(unsigned int index)
{
  // used by doubleclick - calls editor for field, and changes field
  // and return change flag...
  if (ValidPropertyIndex(index))
  {
    switch (c_property[index].type)
    {
      case TYPE_BOOLEAN: // bool - no editor - simply toggle flag
        m_value[index]=( m_value[index] ? 0 : 1 );
        return true;
      case TYPE_NUMBER:
      case TYPE_CHARPTR:
      default:
        break;
    }
  }
  return false;
}
//------------------------------------------------------------------------------
void CBaseObject::Set(unsigned int propertyIndex,const char*ptr)
{
  // set property of object
  if ( !ValidPropertyIndex(propertyIndex) )
    return;
  if ( TYPE_CHARPTR!=c_property[propertyIndex].type )
    Set(propertyIndex,atol(ptr));
  else
  {
    char *ptr2=(char*)m_value[propertyIndex];
    delete[] ptr2;
    ptr2=new char[strlen(ptr)+1];
    strcpy(ptr2,ptr);
    m_value[propertyIndex]=(long)ptr2;
  }
}
//------------------------------------------------------------------------------
long CBaseObject::GetValue(unsigned int i)
{
  // get property value as long
  if (ValidPropertyIndex(i))
  {
    switch (c_property[i].type)
    {
      case TYPE_NUMBER:
      case TYPE_BOOLEAN:
        return m_value[i];
      default:
        break;
    }
  }
  return 0;
}
//------------------------------------------------------------------------------
const char *CBaseObject::GetText(unsigned int i)
{
  // return string value of info - good only for current invocation,
  // so copy value before calling again
  if (ValidPropertyIndex(i))
  {
    switch (c_property[i].type)
    {
      case TYPE_NUMBER:
        wsprintf(s_tempBuff,"%d",m_value[i]);
        return s_tempBuff;
      case TYPE_CHARPTR:
      {
        char *ptr=(char*)m_value[i];
        if (NULL!=ptr)
          return ptr;
        break;
      }
      case TYPE_BOOLEAN:
        strcpy(s_tempBuff,(m_value[i]?"True":"False"));
        return s_tempBuff;
      case TYPE_NONE:
      default:
        break;
    }
  }
  return "";
}
//------------------------------------------------------------------------------
const char *CBaseObject::PropertyName(unsigned int propertyIndex)
{
  // name of indicated property
  if (ValidPropertyIndex(propertyIndex))
    return c_property[propertyIndex].name;
  return "";
}
//------------------------------------------------------------------------------
int CBaseObject::WriteEntry(unsigned int i,HANDLE hFile)
{
  // write property entry to opened file
  DWORD bytesWritten=0, len=0, temp=0;
  BYTE *ptr=NULL;
  switch (c_property[i].type)
  {
    case TYPE_NUMBER:
      len=4;
      temp=m_value[i]&0xFFFFFFFFL;
      if ( 0==WriteFile(hFile,&temp,len,&bytesWritten,NULL) || bytesWritten!=len )
        return 0;
      break;
    case TYPE_BOOLEAN:
      len=1;
      temp=m_value[i]&0x00FFL;
      if ( 0==WriteFile(hFile,&temp,len,&bytesWritten,NULL) || bytesWritten!=len )
        return 0;
      break;
    case TYPE_CHARPTR:
    {
      ptr=(char*)m_value[i];
      if (NULL==ptr)
        ptr="";
      len=strlen(ptr)+1;
      if ( 0==WriteFile(hFile,ptr,len,&bytesWritten,NULL) || bytesWritten!=len )
        return 0;
    }
      break;
    case TYPE_NONE:
    default:
      return 0;
  }
  return len;
}
//------------------------------------------------------------------------------
int CBaseObject::WriteTables(HANDLE hFile)
{
  // write table info to already-open file, and return total bytes written
  // tables depend on table types as to what is written out...
  int len=0, i=0, k=0, totalLen=0, type[]={ TYPE_NUMBER, TYPE_BOOLEAN, TYPE_CHARPTR };
  // write object header type
  DWORD bytesProcessed=0;
  BYTE objType=ObjectType();
  if ( 0==WriteFile(hFile,&objType,1,&bytesProcessed,NULL) || bytesProcessed!=1 )
    return 0;
  for (k=0;k<sizeof(type)/sizeof(type[0]);k++)
  {
    for (i=0;i<TotalProperty();i++)
    {
      if (type[k]==c_property[i].type)
      {
        if (0==(len=WriteEntry(i,hFile)))
          return 0;
        totalLen+=len;
      }
    }
  }
  return totalLen;
}
//------------------------------------------------------------------------------
bool CBaseObject::ReadEntry(unsigned int i,HANDLE hFile)
{
  // read data from open file into tables - return success or false if errors
  DWORD bytesRead=0, temp=0, k=0;
  char buff[256]="", ch='\0';
  switch (c_property[i].type)
  {
    case TYPE_BOOLEAN:
      if ( 0==ReadFile(hFile,&temp,1,&bytesRead,NULL) || bytesRead!=1 )
        return false;
      m_value[i]=temp;
      break;
    case TYPE_NUMBER:
      if ( 0==ReadFile(hFile,&temp,4,&bytesRead,NULL) || bytesRead!=4 )
        return false;
      m_value[i]=temp;
      break;
    case TYPE_CHARPTR:
      k=0;
      do
      {
        if ( 0==ReadFile(hFile,&ch,1,&bytesRead,NULL) || bytesRead!=1 )
          return false;
        buff[k]=ch;
      } while ( ch && ++k<sizeof(buff) );
      if (ch)
        return false;
      Set(i,buff);
      break;
    case TYPE_NONE:
    default:


      break;
  }
  return true;
}
//------------------------------------------------------------------------------
bool CBaseObject::ReadTables(HANDLE hFile)
{
  // read table info from already-open file, and return success
  int len=0, i=0, k=0, totalLen=0, type[]={ TYPE_NUMBER, TYPE_BOOLEAN, TYPE_CHARPTR };
  for (k=0;k<sizeof(type)/sizeof(type[0]);k++)
  {
    for (i=0;i<TotalProperty();i++)
    {
      if (type[k]==c_property[i].type)
      {
        if (!ReadEntry(i,hFile))
          return false;
      }
    }
  }
  return true;
}
//------------------------------------------------------------------------------
void CBaseObject::Draw(TCanvas *canvas,int offsetX,int offsetY)
{
  // do common work before letting each object draw itself
  int left=0, top=0, width=0, height=0;
  GetBounds(&left,&top,&width,&height);
  ObjectDraw(canvas,left+offsetX,top+offsetY,width,height);
}
//---------------------------------------------------------------------------
void CBaseObject::ObjectDraw(TCanvas *canvas,int left,int top,int width,int height)
{
  // base draw - overridden by children
  Windows::TRect rect(left,top,left+width,top+height);
  canvas->Brush->Color=clAqua;
  canvas->Font->Color=clWhite;
  canvas->FillRect(rect);
}
//---------------------------------------------------------------------------
const char *CBaseObject::ClassConstructorString(char *buff)
{
  // for source construction - the constructor id of object
  ObjectClassConstructor(buff);
  return buff;
}
/*===========================================================================


       CHILDREN CLASSES
       ----------------

       CObjButton class


===========================================================================*/
const s_property CObjButton::s_propertyTable[] =
    { { INDEX_CAPTION,"Caption",TYPE_CHARPTR,    0,   0 },
      { INDEX_CLASS,  "Class Name",TYPE_CHARPTR, 3,   0 },
      { INDEX_HEIGHT, "Height", TYPE_NUMBER,    10, 300 },
      { INDEX_LEFT,   "Left",   TYPE_NUMBER, -1000,1000 },
      { INDEX_NAME,   "Name",   TYPE_CHARPTR,    3,   0 },
      { INDEX_TOP,    "Top",    TYPE_NUMBER, -1000,1000 },
      { INDEX_WIDTH,  "Width",  TYPE_NUMBER,    10,3000 } };
//---------------------------------------------------------------------------
CObjButton::CObjButton(void)
 : CBaseObject("CButton",true,OBJ_BUTTON,s_propertyTable,sizeof(s_propertyTable)/sizeof(s_propertyTable[0]),INDEX_LEFT,INDEX_TOP,INDEX_WIDTH,INDEX_HEIGHT,INDEX_NAME,INDEX_CAPTION,INDEX_CLASS),
   m_calcWidth(0)
{
}
//---------------------------------------------------------------------------
void CObjButton::GetBounds(int *left,int *top,int *width,int *height)
{
  // special case: if width=0, button size is determined by text length
  CBaseObject::GetBounds(left,top,width,height);
  if ( 0 == *width )
    *width=m_calcWidth;
}
//------------------------------------------------------------------------------
int CObjButton::CalcWidth(TCanvas *canvas)
{
  // calc width based on text size - use our ebm font for approximate sizing
  AnsiString oldName=canvas->Font->Name;
  int oldSize=canvas->Font->Size;
  TFontStyles oldStyle=canvas->Font->Style;
  canvas->Font->Name=c_defaultFontName;
  canvas->Font->Size=c_defaultFontSize;
  canvas->Font->Style=TFontStyles()<< fsBold;
  const char *text=GetText(INDEX_CAPTION);
  m_calcWidth=canvas->TextWidth(text)+10;
  canvas->Font->Style=oldStyle;
  canvas->Font->Size=oldSize;
  canvas->Font->Name=oldName;
  return m_calcWidth;
}
//------------------------------------------------------------------------------
void CObjButton::ObjectDraw(TCanvas *canvas,int left,int top,int width,int height)
{
  // draw our button shape, with text on it 
  canvas->Brush->Color=c_color[12];
  canvas->Pen->Color=c_color[0];
  canvas->Pen->Width=1;
  canvas->Pen->Style=psSolid;
  canvas->Rectangle(left,top,left+width,top+height);
  canvas->Font->Color=c_color[0];
  canvas->Font->Name=c_defaultFontName;
  canvas->Font->Size=c_defaultFontSize;
  const char *text=GetText(INDEX_CAPTION);
  canvas->Font->Style=TFontStyles()<< fsBold;
  if (0==width)
    width=CalcWidth(canvas);
  int offsetX=(width-canvas->TextWidth(text))/2, offsetY=(height-canvas->TextHeight(text))/2;
  Windows::TRect rect(left+1,top+1,left+width-1,top+height-1);
  canvas->TextRect(rect,left+offsetX,top+offsetY,text);
  canvas->Font->Style=TFontStyles();
}
//---------------------------------------------------------------------------
void CObjButton::ObjectClassConstructor(char *buff)
{
  // output constructor info string
  wsprintf(buff,"%d,%d,\"%s\"",GetValue(INDEX_WIDTH),GetValue(INDEX_HEIGHT),GetText(INDEX_CAPTION));
}
/*===========================================================================


       CObjRepeatButton class


===========================================================================*/
const s_property CObjRepeatButton::s_propertyTable[] =
    { { INDEX_CAPTION,"Caption",TYPE_CHARPTR,    0,   0 },
      { INDEX_CLASS,  "Class Name",TYPE_CHARPTR, 3,   0 },
      { INDEX_HEIGHT, "Height", TYPE_NUMBER,    10, 300 },
      { INDEX_LEFT,   "Left",   TYPE_NUMBER, -1000,1000 },
      { INDEX_NAME,   "Name",   TYPE_CHARPTR,    3,   0 },
      { INDEX_TOP,    "Top",    TYPE_NUMBER, -1000,1000 },
      { INDEX_WIDTH,  "Width",  TYPE_NUMBER,    10,3000 } };
//---------------------------------------------------------------------------
CObjRepeatButton::CObjRepeatButton(void)
 : CBaseObject("CRepeatButton",true,OBJ_REPEATBUTTON,s_propertyTable,sizeof(s_propertyTable)/sizeof(s_propertyTable[0]),INDEX_LEFT,INDEX_TOP,INDEX_WIDTH,INDEX_HEIGHT,INDEX_NAME,INDEX_CAPTION,INDEX_CLASS)
{
}
//---------------------------------------------------------------------------
void CObjRepeatButton::ObjectDraw(TCanvas *canvas,int left,int top,int width,int height)
{
  // draw object
  canvas->Brush->Color=c_color[12];
  canvas->Pen->Color=c_color[0];
  canvas->Pen->Width=1;
  canvas->Pen->Style=psSolid;
  canvas->Rectangle(left,top,left+width,top+height);
  canvas->Font->Color=c_color[0];
  canvas->Font->Name=c_defaultFontName;
  canvas->Font->Size=c_defaultFontSize;
  const char *text=GetText(INDEX_CAPTION);
  canvas->Font->Style=TFontStyles()<< fsBold;
  int offsetX=(width-canvas->TextWidth(text))/2, offsetY=(height-canvas->TextHeight(text))/2;
  Windows::TRect rect(left+1,top+1,left+width-1,top+height-1);
  canvas->TextRect(rect,left+offsetX,top+offsetY,text);
  canvas->Font->Style=TFontStyles();
}
//---------------------------------------------------------------------------
void CObjRepeatButton::ObjectClassConstructor(char *buff)
{
  // output constructor info string
  wsprintf(buff,"%d,%d,\"%s\"",GetValue(INDEX_WIDTH),GetValue(INDEX_HEIGHT),GetText(INDEX_CAPTION));
}
/*===========================================================================


       CObjPushButton class


===========================================================================*/
const s_property CObjPushButton::s_propertyTable[] =
    { { INDEX_CAPTION,"Caption",TYPE_CHARPTR,    0,   0 },
      { INDEX_CLASS,  "Class Name",TYPE_CHARPTR, 3,   0 },
      { INDEX_HEIGHT, "Height", TYPE_NUMBER,    10, 300 },
      { INDEX_LEFT,   "Left",   TYPE_NUMBER, -1000,1000 },
      { INDEX_NAME,   "Name",   TYPE_CHARPTR,    3,   0 },
      { INDEX_TOP,    "Top",    TYPE_NUMBER, -1000,1000 },
      { INDEX_WIDTH,  "Width",  TYPE_NUMBER,    10,3000 } };
//---------------------------------------------------------------------------
CObjPushButton::CObjPushButton(void)
 : CBaseObject("CPushButton",true,OBJ_PUSHBUTTON,s_propertyTable,sizeof(s_propertyTable)/sizeof(s_propertyTable[0]),INDEX_LEFT,INDEX_TOP,INDEX_WIDTH,INDEX_HEIGHT,INDEX_NAME,INDEX_CAPTION,INDEX_CLASS)
{
}
//---------------------------------------------------------------------------
void CObjPushButton::ObjectDraw(TCanvas *canvas,int left,int top,int width,int height)
{
  // draw object
  canvas->Brush->Color=c_color[12];
  canvas->Pen->Color=c_color[0];
  canvas->Pen->Width=1;
  canvas->Pen->Style=psSolid;
  canvas->Rectangle(left,top,left+width,top+height);
  canvas->Font->Color=c_color[0];
  canvas->Font->Name=c_defaultFontName;
  canvas->Font->Size=c_defaultFontSize;
  const char *text=GetText(INDEX_CAPTION);
  canvas->Font->Style=TFontStyles()<< fsBold;
  int offsetX=(width-canvas->TextWidth(text))/2, offsetY=(height-canvas->TextHeight(text))/2;
  Windows::TRect rect(left+1,top+1,left+width-1,top+height-1);
  canvas->TextRect(rect,left+offsetX,top+offsetY,text);
  canvas->Font->Style=TFontStyles();
}
//---------------------------------------------------------------------------
void CObjPushButton::ObjectClassConstructor(char *buff)
{
  // output constructor info string
  wsprintf(buff,"%d,%d,\"%s\"",GetValue(INDEX_WIDTH),GetValue(INDEX_HEIGHT),GetText(INDEX_CAPTION));
}
/*===========================================================================


       CObjCheckBox class


===========================================================================*/
const s_property CObjCheckBox::s_propertyTable[] =
    { { INDEX_CAPTION,"Caption",TYPE_CHARPTR,    0,   0 },
      { INDEX_CLASS,  "Class Name",TYPE_CHARPTR, 3,   0 },
      { INDEX_HEIGHT, "Height", TYPE_NUMBER,    13,  13 },
      { INDEX_LEFT,   "Left",   TYPE_NUMBER, -1000,1000 },
      { INDEX_NAME,   "Name",   TYPE_CHARPTR,    3,   0 },
      { INDEX_TOP,    "Top",    TYPE_NUMBER, -1000,1000 },
      { INDEX_WIDTH,  "Width",  TYPE_NUMBER,    15, 300 } };
//---------------------------------------------------------------------------
CObjCheckBox::CObjCheckBox(void)
 : CBaseObject("CCheckbox",true,OBJ_CHECKBOX,s_propertyTable,sizeof(s_propertyTable)/sizeof(s_propertyTable[0]),INDEX_LEFT,INDEX_TOP,INDEX_WIDTH,INDEX_HEIGHT,INDEX_NAME,INDEX_CAPTION,INDEX_CLASS)
{
}
//---------------------------------------------------------------------------
void CObjCheckBox::ObjectDraw(TCanvas *canvas,int left,int top,int width,int height)
{
  // draw object
  canvas->Pen->Color=clBlack;
  canvas->Pen->Style=psSolid;
  canvas->Brush->Color=clWhite;
  canvas->Rectangle(left,top,left+11,top+11);
  canvas->Font->Name=c_defaultFontName;
  canvas->Font->Size=c_defaultFontSize;
  canvas->Font->Color=c_color[0];
  canvas->Brush->Color=c_color[15];
  canvas->Font->Style=TFontStyles()<< fsBold;
  const char *text=GetText(INDEX_CAPTION);
  int dx=canvas->TextWidth(text), dy=canvas->TextHeight(text);
  if (GetValue(INDEX_WIDTH)>14)
  {
    Windows::TRect rect(left+14,top,left+min(dx+14,(int)GetValue(INDEX_WIDTH)),top+min(dy,(int)GetValue(INDEX_HEIGHT)));
    canvas->TextRect(rect,left+14,top,text);
  }
  canvas->Font->Style=TFontStyles();
}
//---------------------------------------------------------------------------
void CObjCheckBox::ObjectClassConstructor(char *buff)
{
  // output constructor info string
  wsprintf(buff,"%d,%d,\"%s\"",GetValue(INDEX_WIDTH),GetValue(INDEX_HEIGHT),GetText(INDEX_CAPTION));
}
/*===========================================================================


       CObjRadioButton class


===========================================================================*/
const s_property CObjRadioButton::s_propertyTable[] =
    { { INDEX_CAPTION,"Caption",TYPE_CHARPTR,    0,   0 },
      { INDEX_CLASS,  "Class Name",TYPE_CHARPTR, 3,   0 },
      { INDEX_HEIGHT, "Height", TYPE_NUMBER,    13,  13 },
      { INDEX_LEFT,   "Left",   TYPE_NUMBER, -1000,1000 },
      { INDEX_NAME,   "Name",   TYPE_CHARPTR,    3,   0 },
      { INDEX_TOP,    "Top",    TYPE_NUMBER, -1000,1000 },
      { INDEX_WIDTH,  "Width",  TYPE_NUMBER,    15, 300 } };
//---------------------------------------------------------------------------
CObjRadioButton::CObjRadioButton(void)
 : CBaseObject("CRadioButton",true,OBJ_RADIOBUTTON,s_propertyTable,sizeof(s_propertyTable)/sizeof(s_propertyTable[0]),INDEX_LEFT,INDEX_TOP,INDEX_WIDTH,INDEX_HEIGHT,INDEX_NAME,INDEX_CAPTION,INDEX_CLASS)
{
}
//---------------------------------------------------------------------------
void CObjRadioButton::ObjectDraw(TCanvas *canvas,int left,int top,int width,int height)
{
  // draw object
  canvas->Pen->Color=clBlack;
  canvas->Pen->Style=psSolid;
  canvas->Brush->Color=clWhite;
  canvas->Ellipse(left,top,left+11,top+11);
  canvas->Brush->Color=clWhite;
  {
    canvas->Font->Name=c_defaultFontName;
    canvas->Font->Size=c_defaultFontSize;
    canvas->Font->Color=c_color[0];
    canvas->Brush->Color=c_color[15];
    canvas->Font->Style=TFontStyles()<< fsBold;
    const char *text=GetText(INDEX_CAPTION);
    int dx=canvas->TextWidth(text), dy=canvas->TextHeight(text);
    if (GetValue(INDEX_WIDTH)>14)
    {
      Windows::TRect rect(left+14,top,left+min(dx+14,(int)GetValue(INDEX_WIDTH)),top+min(dy,(int)GetValue(INDEX_HEIGHT)));
      canvas->TextRect(rect,left+14,top,text);
    }
    canvas->Font->Style=TFontStyles();
  }
}
//---------------------------------------------------------------------------
void CObjRadioButton::ObjectClassConstructor(char *buff)
{
  // output constructor info string
  wsprintf(buff,"%d,%d,\"%s\"",GetValue(INDEX_WIDTH),GetValue(INDEX_HEIGHT),GetText(INDEX_CAPTION));
}
/*===========================================================================


       CObjTextEdit class


===========================================================================*/
const s_property CObjTextEdit::s_propertyTable[] =
    { { INDEX_CLASS,  "Class Name",TYPE_CHARPTR, 3,   0 },
      { INDEX_HEIGHT,       "Height",            TYPE_NUMBER,    10, 300 },
      { INDEX_LEFT,         "Left",              TYPE_NUMBER, -1000,1000 },
      { INDEX_NAME,         "Name",              TYPE_CHARPTR,    3,   0 },
      { INDEX_S_NOEDIT,     "Set: No Edit",      TYPE_BOOLEAN,    0,   0 },
      { INDEX_S_NOUNDERLINE,"Set: No Underline", TYPE_BOOLEAN,    0,   0 },
      { INDEX_S_ONELINE,    "Set: One Line",     TYPE_BOOLEAN,    0,   0 },
      { INDEX_S_PASSWORD,   "Set: Password",     TYPE_BOOLEAN,    0,   0 },
      { INDEX_TOP,          "Top",               TYPE_NUMBER, -1000,1000 },
      { INDEX_WIDTH,        "Width",             TYPE_NUMBER,    10,3000 } };
//----------------------------------------------------------------------------
CObjTextEdit::CObjTextEdit(void)
 : CBaseObject("CTextEdit",true,OBJ_TEXTEDIT,s_propertyTable,sizeof(s_propertyTable)/sizeof(s_propertyTable[0]),INDEX_LEFT,INDEX_TOP,INDEX_WIDTH,INDEX_HEIGHT,INDEX_NAME,-1,INDEX_CLASS)
{
}
//---------------------------------------------------------------------------
void CObjTextEdit::Set(unsigned int propertyIndex,long value)
{
  CBaseObject::Set(propertyIndex,value);
}
//---------------------------------------------------------------------------
void CObjTextEdit::ObjectDraw(TCanvas *canvas,int left,int top,int width,int height)
{
  // draw object
  Windows::TRect rect(left,top,left+width,top+height);
  canvas->Brush->Color=c_color[15];
  canvas->FillRect(rect);
  canvas->Pen->Color=c_color[0];
  canvas->Pen->Style=psDot;
  int y=top+10;
  while (y<top+height) // draw dotted lines
  {
    canvas->MoveTo(left,y);
    canvas->LineTo(left+width,y);
    y+=12;
  }
}
//---------------------------------------------------------------------------
void CObjTextEdit::ObjectClassConstructor(char *buff)
{
  // output constructor info string
  wsprintf(buff,"%d,%d,",GetValue(INDEX_WIDTH),GetValue(INDEX_HEIGHT));
  // add options, seperated by '|' operator
  struct
  {
    int index;
    char *text;
  } data[]={ { INDEX_S_NOEDIT,     "TEXTOPTION_NOEDIT",      },
             { INDEX_S_NOUNDERLINE,"TEXTOPTION_NOUNDERLINE", },
             { INDEX_S_ONELINE,    "TEXTOPTION_ONELINE",     },
             { INDEX_S_PASSWORD,   "TEXTOPTION_PASSWORD"     } };
  bool first=true;
  for (int i=0;i<sizeof(data)/sizeof(data[0]);i++)
  {
    if (GetValue(data[i].index))
    {
      if (!first)
        strcat(buff,"|");
      strcat(buff,data[i].text);
      first=false;
    }
  }
  if (first)
   strcat(buff,"0");
}
/*===========================================================================


       CObjScrollBar class


===========================================================================*/
const s_property CObjScrollBar::s_propertyTable[] =
    { { INDEX_CLASS,  "Class Name",TYPE_CHARPTR, 3,   0 },
      { INDEX_HORIZONTAL,"Horizontal",TYPE_BOOLEAN,    0,   0 },
      { INDEX_LEFT,      "Left",      TYPE_NUMBER, -1000,1000 },
      { INDEX_NAME,      "Name",      TYPE_CHARPTR,    3,   0 },
      { INDEX_SIZE,      "Size",      TYPE_NUMBER,    26, 300 },
      { INDEX_TOP,       "Top",       TYPE_NUMBER, -1000,1000 } };
//---------------------------------------------------------------------------
CObjScrollBar::CObjScrollBar(void)
 : CBaseObject("CScrollBar",true,OBJ_SCROLLBAR,s_propertyTable,sizeof(s_propertyTable)/sizeof(s_propertyTable[0]),INDEX_LEFT,INDEX_TOP,INDEX_WIDTH,INDEX_HEIGHT,INDEX_NAME,-1,INDEX_CLASS)
{
  Set(INDEX_HORIZONTAL,1);
}
//---------------------------------------------------------------------------
long CObjScrollBar::GetValue(unsigned int propertyIndex)
{
  // special case - horizontal scrolls always have a fixed vertical size,
  // and reverse for vertical ones - this call checks and enforces that
  switch (propertyIndex)
  {
    case INDEX_HEIGHT:
      if (GetValue(INDEX_HORIZONTAL))
        return 12;
      return CBaseObject::GetValue(INDEX_SIZE);
    case INDEX_WIDTH:
      if (!GetValue(INDEX_HORIZONTAL))
        return 12;
      return CBaseObject::GetValue(INDEX_SIZE);
  }
  return CBaseObject::GetValue(propertyIndex);
}
//---------------------------------------------------------------------------
void CObjScrollBar::Set(unsigned int propertyIndex,long value)
{
  // check/verify sizes for scrollbars 
  switch (propertyIndex)
  {
    case INDEX_HEIGHT:
      if (!GetValue(INDEX_HORIZONTAL))
        CBaseObject::Set(INDEX_SIZE,max( s_propertyTable[INDEX_SIZE].param1, min( value, s_propertyTable[INDEX_SIZE].param2 )));
      return;
    case INDEX_WIDTH:
      if (GetValue(INDEX_HORIZONTAL))
        CBaseObject::Set(INDEX_SIZE,max( s_propertyTable[INDEX_SIZE].param1, min( value, s_propertyTable[INDEX_SIZE].param2 )));
      return;
  }
  CBaseObject::Set(propertyIndex,value);
}
//---------------------------------------------------------------------------
void CObjScrollBar::GetBounds(int *left,int *top,int *width,int *height)
{
  // special override - get specific values for size
  *left  =GetValue(INDEX_LEFT);
  *top   =GetValue(INDEX_TOP);
  *width =GetValue(INDEX_WIDTH);
  *height=GetValue(INDEX_HEIGHT);
}
//------------------------------------------------------------------------------
void CObjScrollBar::SetBounds(long left,long top,long width,long height)
{
  // set values
  Set(INDEX_LEFT,  left);
  Set(INDEX_TOP,   top);
  Set(INDEX_WIDTH, width);
  Set(INDEX_HEIGHT,height);
}
//------------------------------------------------------------------------------
void CObjScrollBar::ObjectDraw(TCanvas *canvas,int left,int top,int width,int height)
{
  // draw object
  canvas->Pen->Color=c_color[0];
  canvas->Pen->Style=psSolid;
  canvas->Brush->Color=c_color[10];
  Windows::TRect rect;
  Windows::TPoint points[4];
  if (GetValue(INDEX_HORIZONTAL))
  {
    canvas->Rectangle(left,top,left+width,top+12);
    canvas->Brush->Color=c_color[0];
    Windows::TRect rect1(left,top,left+12,top+12);
    canvas->FillRect(rect1);
    Windows::TRect rect2(left+width-12,top,left+width,top+12);
    canvas->FillRect(rect2);
    // now arrows
    points[0] = Point(left,top+5);
    points[1] = Point(left+10,top);
    points[2] = Point(left+10,top+11);
    canvas->Brush->Color=c_color[11];
    canvas->Polygon(points, 2);
    points[0] = Point(left+width-1,top+5);
    points[1] = Point(left+width-11,top);
    points[2] = Point(left+width-11,top+11);
    canvas->Brush->Color=c_color[11];
    canvas->Polygon(points, 2);
  }
  else // vertical
  {
    canvas->Rectangle(left,top,left+12,top+height);
    canvas->Brush->Color=c_color[0];
    Windows::TRect rect1(left,top,left+12,top+12);
    canvas->FillRect(rect1);
    Windows::TRect rect2(left,top+height-12,left+12,top+height);
    canvas->FillRect(rect2);
    // now arrows
    points[0] = Point(left+5,top);
    points[1] = Point(left+11,top+10);
    points[2] = Point(left,top+10);
    canvas->Brush->Color=c_color[11];
    canvas->Polygon(points, 2);
    points[0] = Point(left+5,top+height-1);
    points[1] = Point(left+11,top+height-11);
    points[2] = Point(left,top+height-11);
    canvas->Brush->Color=c_color[11];
    canvas->Polygon(points, 2);
  }
}
//---------------------------------------------------------------------------
void CObjScrollBar::ObjectClassConstructor(char *buff)
{
  // output constructor info string
  if (GetValue(INDEX_HORIZONTAL))
    wsprintf(buff,"%d,%d",GetValue(INDEX_WIDTH),0);
  else
    wsprintf(buff,"%d,%d",0,GetValue(INDEX_HEIGHT));
}
/*===========================================================================


       CObjProgressBar class


===========================================================================*/
const s_property CObjProgressBar::s_propertyTable[] =
    { { INDEX_CLASS,  "Class Name",TYPE_CHARPTR, 3,   0 },
      { INDEX_HORIZONTAL,"Horizontal",TYPE_BOOLEAN,    0,   0 },
      { INDEX_LEFT,      "Left",      TYPE_NUMBER, -1000,1000 },
      { INDEX_NAME,      "Name",      TYPE_CHARPTR,    3,   0 },
      { INDEX_SIZE,      "Size",      TYPE_NUMBER,    26, 300 },
      { INDEX_TOP,       "Top",       TYPE_NUMBER, -1000,1000 } };
//---------------------------------------------------------------------------
CObjProgressBar::CObjProgressBar(void)
 : CBaseObject("CProgressBar",true,OBJ_PROGRESSBAR,s_propertyTable,sizeof(s_propertyTable)/sizeof(s_propertyTable[0]),INDEX_LEFT,INDEX_TOP,INDEX_WIDTH,INDEX_HEIGHT,INDEX_NAME,-1,INDEX_CLASS)
{
  // by default, horizontal
  Set(INDEX_HORIZONTAL,1);
}
//---------------------------------------------------------------------------
long CObjProgressBar::GetValue(unsigned int propertyIndex)
{
  // clamp values, much like scrollbars
  switch (propertyIndex)
  {
    case INDEX_HEIGHT:
      if (GetValue(INDEX_HORIZONTAL))
        return 12;
      return GetValue(INDEX_SIZE);
    case INDEX_WIDTH:
      if (!GetValue(INDEX_HORIZONTAL))
        return 12;
      return GetValue(INDEX_SIZE);
  }
  return CBaseObject::GetValue(propertyIndex);
}
//---------------------------------------------------------------------------
void CObjProgressBar::Set(unsigned int propertyIndex,long value)
{
  // clamp values, much like scrollbars
  switch (propertyIndex)
  {
    case INDEX_HEIGHT:
      if (!GetValue(INDEX_HORIZONTAL))
        Set(INDEX_SIZE,max( s_propertyTable[INDEX_SIZE].param1, min( value, s_propertyTable[INDEX_SIZE].param2 )));
      return;
    case INDEX_WIDTH:
      if (GetValue(INDEX_HORIZONTAL))
        Set(INDEX_SIZE,max( s_propertyTable[INDEX_SIZE].param1, min( value, s_propertyTable[INDEX_SIZE].param2 )));
      return;
  }
  CBaseObject::Set(propertyIndex,value);
}
//---------------------------------------------------------------------------
void CObjProgressBar::GetBounds(int *left,int *top,int *width,int *height)
{
  *left  =GetValue(INDEX_LEFT);
  *top   =GetValue(INDEX_TOP);
  *width =GetValue(INDEX_WIDTH);
  *height=GetValue(INDEX_HEIGHT);
}
//------------------------------------------------------------------------------
void CObjProgressBar::SetBounds(long left,long top,long width,long height)
{
  Set(INDEX_LEFT,  left);
  Set(INDEX_TOP,   top);
  Set(INDEX_WIDTH, width);
  Set(INDEX_HEIGHT,height);
}
//------------------------------------------------------------------------------
void CObjProgressBar::ObjectDraw(TCanvas *canvas,int left,int top,int width,int height)
{
  // draw object
  canvas->Pen->Color=c_color[0];
  canvas->Pen->Style=psSolid;
  canvas->Brush->Color=c_color[15];
  Windows::TRect rect;
  if (GetValue(INDEX_HORIZONTAL))
    canvas->Rectangle(left,top,left+width,top+12);
  else
    canvas->Rectangle(left,top,left+12,top+height);
}
//---------------------------------------------------------------------------
void CObjProgressBar::ObjectClassConstructor(char *buff)
{
  // output constructor info string
  if (GetValue(INDEX_HORIZONTAL))
    wsprintf(buff,"%d,%d",GetValue(INDEX_WIDTH),0);
  else
    wsprintf(buff,"%d,%d",0,GetValue(INDEX_HEIGHT));
}
/*===========================================================================


       CObjLabel class


===========================================================================*/
const s_property CObjLabel::s_propertyTable[] =
    { { INDEX_CAPTION,  "Caption",         TYPE_CHARPTR,   0,    0 },
      { INDEX_CLASS,    "Class Name",      TYPE_CHARPTR,   3,    0 },
      { INDEX_HEIGHT,   "Height",          TYPE_NUMBER,    13,  13 },
      { INDEX_LEFT,     "Left",            TYPE_NUMBER, -1000,1000 },
      { INDEX_NAME,     "Name",            TYPE_CHARPTR,    3,   0 },
      { INDEX_TOP,      "Top",             TYPE_NUMBER, -1000,1000 },
      { INDEX_WIDTH,    "Width",           TYPE_NUMBER,    10,3000 } };
//---------------------------------------------------------------------------
CObjLabel::CObjLabel(void)
 : CBaseObject("CLabel",true,OBJ_LABEL,s_propertyTable,sizeof(s_propertyTable)/sizeof(s_propertyTable[0]),INDEX_LEFT,INDEX_TOP,INDEX_WIDTH,INDEX_HEIGHT,INDEX_NAME,INDEX_CAPTION,INDEX_CLASS)
{
}
//---------------------------------------------------------------------------
void CObjLabel::ObjectDraw(TCanvas *canvas,int left,int top,int width,int height)
{
  canvas->Brush->Color=c_color[15];
  canvas->Font->Color=c_color[0];
  canvas->Font->Name=c_defaultFontName;
  canvas->Font->Size=c_defaultFontSize;
  const char *text=GetText(INDEX_CAPTION);
  int dx=canvas->TextWidth(text), dy=canvas->TextHeight(text);
  Windows::TRect rect(left,top,min(left+width,left+dx),top+dy);
  canvas->TextRect(rect,rect.left,rect.top,text);
}
//---------------------------------------------------------------------------
void CObjLabel::ObjectClassConstructor(char *buff)
{
  // output constructor info string
  wsprintf(buff,"\"%s\",0",GetText(INDEX_CAPTION));
}
/*===========================================================================


       CObjRect class


===========================================================================*/
const s_property CObjRect::s_propertyTable[] =
    { { INDEX_CLASS,    "Class Name",      TYPE_CHARPTR,    3,   0 },
      { INDEX_HEIGHT,   "Height",          TYPE_NUMBER,     1, 400 },
      { INDEX_LEFT,     "Left",            TYPE_NUMBER, -1000,1000 },
      { INDEX_NAME,     "Name",            TYPE_CHARPTR,    3,   0 },
      { INDEX_TOP,      "Top",             TYPE_NUMBER, -1000,1000 },
      { INDEX_WIDTH,    "Width",           TYPE_NUMBER,     1, 400 } };
//---------------------------------------------------------------------------
CObjRect::CObjRect(void)
 : CBaseObject("CRect",true,OBJ_RECT,s_propertyTable,sizeof(s_propertyTable)/sizeof(s_propertyTable[0]),INDEX_LEFT,INDEX_TOP,INDEX_WIDTH,INDEX_HEIGHT,INDEX_NAME,-1,INDEX_CLASS)
{
}
//---------------------------------------------------------------------------
void CObjRect::ObjectDraw(TCanvas *canvas,int left,int top,int width,int height)
{
  // draw object
  Windows::TRect rect(left,top,left+width,top+height);
  canvas->Brush->Color=c_color[0];
  canvas->FillRect(rect);
}
//---------------------------------------------------------------------------
void CObjRect::ObjectClassConstructor(char *buff)
{
  // output constructor info string
  wsprintf(buff,"%d,%d",GetValue(INDEX_WIDTH),GetValue(INDEX_HEIGHT));
}
/*===========================================================================


       CObjIcon class


===========================================================================*/
const s_property CObjIcon::s_propertyTable[] =
    { { INDEX_CLASS,  "Class Name",TYPE_CHARPTR, 3,   0 },
      { INDEX_HEIGHT,   "Height",          TYPE_NUMBER,    10, 300 },
      { INDEX_IMAGENAME,"Image Name",      TYPE_CHARPTR,    0,   0 },
      { INDEX_LEFT,     "Left",            TYPE_NUMBER, -1000,1000 },
      { INDEX_NAME,     "Name",            TYPE_CHARPTR,    3,   0 },
      { INDEX_TOP,      "Top",             TYPE_NUMBER, -1000,1000 },
      { INDEX_WIDTH,    "Width",           TYPE_NUMBER,    10,3000 } };
//---------------------------------------------------------------------------
CObjIcon::CObjIcon(void)
 : CBaseObject("CIcon",true,OBJ_ICON,s_propertyTable,sizeof(s_propertyTable)/sizeof(s_propertyTable[0]),INDEX_LEFT,INDEX_TOP,INDEX_WIDTH,INDEX_HEIGHT,INDEX_NAME,-1,INDEX_CLASS),
   m_image(NULL)
{
}
//---------------------------------------------------------------------------
CObjIcon::~CObjIcon(void)
{
  delete m_image;
}
//---------------------------------------------------------------------------
void CObjIcon::SetBounds(long left,long top,long width,long height)
{
  // prevent setting of height/width/picture flag without interference...
  CBaseObject::SetBounds(left,top,width,height);
}
//---------------------------------------------------------------------------
void CObjIcon::ObjectDraw(TCanvas *canvas,int left,int top,int width,int height)
{
  // draw object
  Windows::TRect rect(left,top,left+width,top+height);
  Windows::TRect src(0,0,width,height);
  canvas->Pen->Color=c_color[2];
  canvas->Pen->Style=psSolid;
  canvas->Brush->Color=c_color[15];
  canvas->FillRect(rect);
  canvas->MoveTo(rect.left,rect.top);
  canvas->LineTo(rect.right+1,rect.bottom+1);
  canvas->MoveTo(rect.right,rect.top);
  canvas->LineTo(rect.left-1,rect.bottom+1);
}
//---------------------------------------------------------------------------
void CObjIcon::ObjectClassConstructor(char *buff)
{
  // output constructor info string
  if (0==strlen(GetText(INDEX_IMAGENAME))) // no entry?
    wsprintf(buff,"(IMAGE*)NULL");
  else
    wsprintf(buff,GetText(INDEX_IMAGENAME));
}
/*===========================================================================


       CObjTable class


===========================================================================*/
const s_property CObjTable::s_propertyTable[] =
    { { INDEX_CLASS,  "Class Name",TYPE_CHARPTR, 3,   0 },
      { INDEX_HEIGHT,   "Height",          TYPE_NUMBER,    10, 300 },
      { INDEX_LEFT,     "Left",            TYPE_NUMBER, -1000,1000 },
      { INDEX_NAME,     "Name",            TYPE_CHARPTR,    3,   0 },
      { INDEX_TOP,      "Top",             TYPE_NUMBER, -1000,1000 },
      { INDEX_WIDTH,    "Width",           TYPE_NUMBER,    10,3000 } };
//---------------------------------------------------------------------------
CObjTable::CObjTable(void)
 : CBaseObject("CTable",true,OBJ_TABLE,s_propertyTable,sizeof(s_propertyTable)/sizeof(s_propertyTable[0]),INDEX_LEFT,INDEX_TOP,INDEX_WIDTH,INDEX_HEIGHT,INDEX_NAME,-1,INDEX_CLASS)
{
}
//---------------------------------------------------------------------------
void CObjTable::ObjectDraw(TCanvas *canvas,int left,int top,int width,int height)
{
  // draw object
  canvas->Pen->Color=c_color[8];
  canvas->Pen->Style=psSolid;
  canvas->Brush->Color=c_color[15];
  canvas->Rectangle(left,top,left+width,top+height);
}
//---------------------------------------------------------------------------
void CObjTable::ObjectClassConstructor(char *buff)
{
  // output constructor info string
  wsprintf(buff,"%d,%d",GetValue(INDEX_WIDTH),GetValue(INDEX_HEIGHT));
}
/*===========================================================================


       CObjDateEdit class


===========================================================================*/
const s_property CObjDateEdit::s_propertyTable[] =
    { { INDEX_CLASS,  "Class Name",TYPE_CHARPTR, 3,   0 },
      { INDEX_HEIGHT,   "Height",          TYPE_NUMBER,    12,  12 },
      { INDEX_LEFT,     "Left",            TYPE_NUMBER, -1000,1000 },
      { INDEX_NAME,     "Name",            TYPE_CHARPTR,    3,   0 },
      { INDEX_TOP,      "Top",             TYPE_NUMBER, -1000,1000 },
      { INDEX_WIDTH,    "Width",           TYPE_NUMBER,    71,  71 } };
//---------------------------------------------------------------------------
CObjDateEdit::CObjDateEdit(void)
 : CBaseObject("CDateEdit",true,OBJ_DATEEDIT,s_propertyTable,sizeof(s_propertyTable)/sizeof(s_propertyTable[0]),INDEX_LEFT,INDEX_TOP,INDEX_WIDTH,INDEX_HEIGHT,INDEX_NAME,-1,INDEX_CLASS)
{
}
//---------------------------------------------------------------------------
void CObjDateEdit::ObjectDraw(TCanvas *canvas,int left,int top,int width,int height)
{
  // draw object
  canvas->Font->Name=c_defaultFontName;
  canvas->Font->Size=c_defaultFontSize;
  canvas->Font->Color=c_color[0];
  canvas->Brush->Color=c_color[15];
  canvas->Font->Style=TFontStyles()<< fsBold;
  canvas->Brush->Color=c_color[15];
  Windows::TRect rect(left,top,left+71,top+12);
  canvas->TextRect(rect,left+9,top,"12/31/01");
  canvas->Font->Style=TFontStyles();
  // now triangle down
  canvas->Pen->Color=c_color[0];
  canvas->Pen->Style=psSolid;
  Windows::TPoint points[3];
  points[0] = Point(left+3,top+10);
  points[1] = Point(left+7,top+1);
  points[2] = Point(left,top+1);
  canvas->Brush->Color=c_color[0];
  canvas->Polygon(points, 2);
}
//---------------------------------------------------------------------------
void CObjDateEdit::ObjectClassConstructor(char *buff)
{
  // output constructor info string
  wsprintf(buff,"%d",GetValue(INDEX_WIDTH),0);
}
/*===========================================================================


       CObjTimeEdit class


===========================================================================*/
const s_property CObjTimeEdit::s_propertyTable[] =
    { { INDEX_CLASS,  "Class Name",TYPE_CHARPTR, 3,   0 },
      { INDEX_HEIGHT,   "Height",          TYPE_NUMBER,    12,  12 },
      { INDEX_LEFT,     "Left",            TYPE_NUMBER, -1000,1000 },
      { INDEX_NAME,     "Name",            TYPE_CHARPTR,    3,   0 },
      { INDEX_TOP,      "Top",             TYPE_NUMBER, -1000,1000 },
      { INDEX_WIDTH,    "Width",           TYPE_NUMBER,    83,  83 } };
//---------------------------------------------------------------------------
CObjTimeEdit::CObjTimeEdit(void)
 : CBaseObject("CTimeEdit",true,OBJ_TIMEEDIT,s_propertyTable,sizeof(s_propertyTable)/sizeof(s_propertyTable[0]),INDEX_LEFT,INDEX_TOP,INDEX_WIDTH,INDEX_HEIGHT,INDEX_NAME,-1,INDEX_CLASS)
{
}
//---------------------------------------------------------------------------
void CObjTimeEdit::ObjectDraw(TCanvas *canvas,int left,int top,int width,int height)
{
  // draw object
  canvas->Font->Name=c_defaultFontName;
  canvas->Font->Size=c_defaultFontSize;
  canvas->Font->Color=c_color[0];
  canvas->Brush->Color=c_color[15];
  canvas->Font->Style=TFontStyles()<< fsBold;
  canvas->Brush->Color=c_color[15];
  Windows::TRect rect(left,top,left+83,top+12);
  canvas->TextRect(rect,left+9,top,"11:");
  rect.left+=40;
  canvas->TextRect(rect,rect.left,top,"45");
  rect.left+=26;
  canvas->TextRect(rect,rect.left,top,"PM");
  canvas->Font->Style=TFontStyles();
  // now triangle down
  canvas->Pen->Color=c_color[0];
  canvas->Pen->Style=psSolid;
  Windows::TPoint points[3];
  canvas->Brush->Color=c_color[0];
  points[0] = Point(left+3,top+10);
  points[1] = Point(left+7,top+1);
  points[2] = Point(left,top+1);
  canvas->Polygon(points, 2);
  points[0] = Point(left+33,top+10);
  points[1] = Point(left+37,top+1);
  points[2] = Point(left+30,top+1);
  canvas->Polygon(points, 2);
  points[0] = Point(left+60,top+10);
  points[1] = Point(left+64,top+1);
  points[2] = Point(left+57,top+1);
  canvas->Polygon(points, 2);
}
//---------------------------------------------------------------------------
void CObjTimeEdit::ObjectClassConstructor(char *buff)
{
  // output constructor info string
  wsprintf(buff,"%d",-1); // current time for now...
}
/*===========================================================================


       CObjForm class


===========================================================================*/
const s_property CObjForm::s_propertyTable[] =
    { { INDEX_CAPTION,    "Caption",          TYPE_CHARPTR,    0,   0 },
      { INDEX_CLASS,      "Class Name",       TYPE_CHARPTR,    3,   0 },
      { INDEX_CLOSEBUTTON,"Close Button",     TYPE_BOOLEAN,    0,   0 },
      { INDEX_ENUM,       "Form Number Base", TYPE_NUMBER,     1,32000 },
      { INDEX_HEIGHT,     "Height",           TYPE_NUMBER,    10, 300 },
      { INDEX_LEFT,       "Left",             TYPE_NUMBER, -1000,1000 },
      { INDEX_NAME,       "Name",             TYPE_CHARPTR,    3,   0 },
      { INDEX_PORTRAIT_O, "Portrait Orientation",TYPE_BOOLEAN,    0,   0 },
      { INDEX_SAVEUNDER,  "Save Under Window",TYPE_BOOLEAN,    0,   0 },
      { INDEX_TITLEBAR,   "Title Bar",        TYPE_BOOLEAN,    0,   0 },
      { INDEX_TOP,        "Top",              TYPE_NUMBER, -1000,1000 },
      { INDEX_VISIBLE,    "Visible",          TYPE_BOOLEAN,    0,   0 },
      { INDEX_WIDTH,      "Width",            TYPE_NUMBER,    10, 300 } };
//---------------------------------------------------------------------------
CObjForm::CObjForm(void)
 : CBaseObject("CWindow",false,OBJ_FORM,s_propertyTable,sizeof(s_propertyTable)/sizeof(s_propertyTable[0]),INDEX_LEFT,INDEX_TOP,INDEX_WIDTH,INDEX_HEIGHT,INDEX_NAME,INDEX_CAPTION,INDEX_CLASS)
{
  Set(INDEX_VISIBLE,1);
  Set(INDEX_CLOSEBUTTON,1);
  Set(INDEX_TITLEBAR,1);
  Set(INDEX_PORTRAIT_O,1);
  Set(INDEX_ENUM,100);
}
//---------------------------------------------------------------------------
void CObjForm::ObjectDraw(TCanvas *canvas,int left,int top,int width,int height)
{
  // draw object
  FormDraw(canvas,left,top,width,height,true);
}
//---------------------------------------------------------------------------
void CObjForm::NonClientDraw(TCanvas *canvas,int offsetX,int offsetY)
{
  // draw just outside of form (non-client area/frame)
  int left=0, top=0, width=0, height=0;
  GetBounds(&left,&top,&width,&height);
  FormDraw(canvas,left+offsetX,top+offsetY,width,height,false);
}
//---------------------------------------------------------------------------
void CObjForm::FormDraw(TCanvas *canvas,int left,int top,int width,int height,bool drawClient)
{
  // draw our form properly on background
  int topEdge=( GetValue(INDEX_TITLEBAR) ? 16 : 2 );
  {
    Windows::TRect rect(left,top,left+width,top+height);
    canvas->Pen->Style=psSolid;
    canvas->Brush->Color=clBlack;
    canvas->Pen->Color=clBlack;
    canvas->FrameRect(rect);
  }
  {
    Windows::TRect rect(left+1,top+topEdge-1,left+width-1,top+height-1);
    canvas->Brush->Color=c_color[8];
    canvas->FrameRect(rect);
  }
  if (drawClient)
  {
    canvas->Pen->Color=c_color[8];
    canvas->Brush->Color=c_color[14];
    canvas->Rectangle(left+1,top+topEdge-1,left+width-1,top+height-1);
  }
  {
    Windows::TRect rect(left,top,left+width,top+topEdge-1);
    canvas->Font->Color=clBlack;
    canvas->Font->Style=TFontStyles()<<fsBold;
    canvas->Font->Color=clWhite;
    canvas->Font->Name=c_defaultFontName;
    canvas->Font->Size=c_defaultFontSize;
    const char *text=GetText(INDEX_CAPTION);
    canvas->Brush->Color=clBlack;
    canvas->TextRect(rect,rect.left+4,rect.top+2,text);
    canvas->Font->Style=TFontStyles()>>fsBold;
  }
  // if close box, then draw it
  if ( GetValue(INDEX_CLOSEBUTTON) && GetValue(INDEX_TITLEBAR) )
  {
    Windows::TRect rect(left+width-13,top+3,left+width-3,top+13);
    canvas->Brush->Color=c_color[15];
    canvas->FillRect(rect);
    canvas->Pen->Color=c_color[0];
    canvas->Pen->Style=psSolid;
    int x1=rect.left+2, x2=rect.right-1;
    int y1=rect.top+1,  y2=rect.bottom-1;
    canvas->MoveTo(x1,y1);
    canvas->LineTo(x2,y2);
    canvas->MoveTo(x2-1,y1);
    canvas->LineTo(x1-1,y2);
    canvas->MoveTo(x1-1,y1);
    canvas->LineTo(x2-1,y2);
    canvas->MoveTo(x2-2,y1);
    canvas->LineTo(x1-2,y2);
  }
}
//---------------------------------------------------------------------------
void CObjForm::ObjectClassConstructor(char *buff)
{
  // no constructor string needed for form - just make sure string empty
  buff[0]='\0';
}
//---------------------------------------------------------------------------

















