#ifndef pabStructH
#define pabStructH
/*===========================================================================


       objcreat.h   - windows imitations of ebm object classes  

                      Copyright  2001, 2003 by Pankhurst Algorithmics
                      (www.pankhurst.com, www.ebmDevMag.com)
                      All Rights Reserved


===========================================================================*/
#ifndef BYTE
typedef unsigned char BYTE;
#endif
//---------------------------------------------------------------------------
enum e_objType // ids of each child object
{
  OBJ_NONE=0,
  OBJ_FORM,
  OBJ_BUTTON,
  OBJ_CHECKBOX,
  OBJ_RADIOBUTTON,
  OBJ_TEXTEDIT,
  OBJ_SCROLLBAR,
  OBJ_LABEL,
  OBJ_TABLE,
  OBJ_REPEATBUTTON,
  OBJ_PUSHBUTTON,
  OBJ_PROGRESSBAR,
  OBJ_RECT,
  OBJ_ICON,
  OBJ_TIMEEDIT,
  OBJ_DATEEDIT,
  OBJ_MAXTYPE // limit checking
};
//------------------------------------------------------------------------------
enum e_paramType // for i/o - indicates data read in/out
{
  TYPE_NONE=0,
  TYPE_BOOLEAN,
  TYPE_NUMBER,
  TYPE_CHARPTR,
};
//------------------------------------------------------------------------------
struct s_property // property struct to describe objects
{
  int  index;
  char *name;
  e_paramType type;
  long param1, param2; // used as needed by type
};
//------------------------------------------------------------------------------
class CBaseObject // virtual base class for all objects
{

  private:

    enum { c_maxId=500 };

    int   m_indexLeft, m_indexTop, m_indexWidth, m_indexHeight;
    int   m_indexName, m_indexCaption, m_indexClass;
    bool  m_needsHandler;
    e_objType m_objType;

    static char s_tempBuff[99];
    static BYTE s_idsInUse[c_maxId];

    bool ValidPropertyIndex(unsigned int index)
    {  return (index<c_totalProperty);   }
    void CleanUpTableValues(void);

  protected:

    const s_property *c_property; // pointer to array created by child class
    const unsigned int c_totalProperty;
    long *m_value; // changed by child - this is flexible...

    virtual void ObjectClassConstructor(char *buff) = 0;

    virtual void ObjectDraw(TCanvas *canvas,int left,int top,int width,int height);
    virtual int  WriteEntry(unsigned int i,HANDLE hFile);
    virtual bool ReadEntry(unsigned int i,HANDLE hFile);

  public:

    CBaseObject(const char *className,bool needsHandler,e_objType objType,
                const s_property *propTable,int propTableLen, // pointer to array passed by child class
                int indexLeft,int indexTop,int indexWidth,
                int indexHeight,int indexName,int indexCaption,int indexClass );
    virtual ~CBaseObject(void);
    const char *GetText(unsigned int propertyIndex);
    long GetValue(unsigned int propertyIndex);
    const char *PropertyName(unsigned int propertyIndex);
    e_paramType Type(unsigned int pIndex)
    {  return ( ValidPropertyIndex(pIndex) ? c_property[pIndex].type : TYPE_NONE );  }
    unsigned int TotalProperty(void)
    {  return c_totalProperty;  }
    void Set(unsigned int propertyIndex,const char *value);
    int  WriteTables(HANDLE hFile);
    bool ReadTables(HANDLE hFile);
    // access
    virtual void GetBounds(int *left,int *top,int *width,int *height);
    const char *GetName(void);
    void SetName(const char*name);
    void Draw(TCanvas *canvas,int offsetX,int offsetY);
    bool TestAndSet(unsigned int propertyIndex,const unsigned char*ptr);
    int  IndexName(void)
    {  return m_indexName;  }
    const char*ClassName(void)
    {  return GetText(m_indexClass);  }
    const char*ClassConstructorString(char *buff);
    int  DefaultName(void);
    bool NeedsHandler(void)
    {  return m_needsHandler; }
    e_objType ObjectType(void)
    {  return m_objType;      }

    virtual bool Editor(unsigned int index);
    virtual void Set(unsigned int propertyIndex,long value);
    virtual void SetBounds(long left,long top,long width,long height);

};
//---------------------------------------------------------------------------
class CObjButton : public CBaseObject
{

  private:

    enum { INDEX_CAPTION, INDEX_CLASS, INDEX_HEIGHT, INDEX_LEFT, INDEX_NAME, INDEX_TOP, INDEX_WIDTH, };

    int  m_calcWidth;

    static const s_property s_propertyTable[];

    void ObjectClassConstructor(char *buff);
    void ObjectDraw(TCanvas *canvas,int left,int top,int width,int height);
    void GetBounds(int *left,int *top,int *width,int *height);
    int  CalcWidth(TCanvas *canvas);

  public:

    CObjButton(void);

};
//---------------------------------------------------------------------------
class CObjRepeatButton : public CBaseObject
{

  private:

    enum { INDEX_CAPTION, INDEX_CLASS, INDEX_HEIGHT, INDEX_LEFT, INDEX_NAME, INDEX_TOP, INDEX_WIDTH, };

    static const s_property s_propertyTable[];

    void ObjectClassConstructor(char *buff);
    void ObjectDraw(TCanvas *canvas,int left,int top,int width,int height);

  public:

    CObjRepeatButton(void);

};
//---------------------------------------------------------------------------
class CObjPushButton : public CBaseObject
{

  private:

    enum { INDEX_CAPTION, INDEX_CLASS, INDEX_HEIGHT, INDEX_LEFT, INDEX_NAME, INDEX_TOP, INDEX_WIDTH, };

    static const s_property s_propertyTable[];

    void ObjectClassConstructor(char *buff);
    void ObjectDraw(TCanvas *canvas,int left,int top,int width,int height);

  public:

    CObjPushButton(void);

};
//---------------------------------------------------------------------------
class CObjCheckBox : public CBaseObject
{

  private:

    enum { INDEX_CAPTION, INDEX_CLASS, INDEX_HEIGHT, INDEX_LEFT, INDEX_NAME, INDEX_TOP, INDEX_WIDTH, };

    static const s_property s_propertyTable[];

    void ObjectClassConstructor(char *buff);
    void ObjectDraw(TCanvas *canvas,int left,int top,int width,int height);

  public:

    CObjCheckBox(void);
};
//---------------------------------------------------------------------------
class CObjRadioButton : public CBaseObject
{

  private:

    enum { INDEX_CAPTION, INDEX_CLASS, INDEX_HEIGHT, INDEX_LEFT, INDEX_NAME, INDEX_TOP, INDEX_WIDTH, };

    static const s_property s_propertyTable[];

    void ObjectClassConstructor(char *buff);
    void ObjectDraw(TCanvas *canvas,int left,int top,int width,int height);

  public:

    CObjRadioButton(void);
};
//---------------------------------------------------------------------------
class CObjTextEdit : public CBaseObject
{

  private:

    enum { INDEX_CLASS, INDEX_HEIGHT, INDEX_LEFT, INDEX_NAME, INDEX_S_NOEDIT, INDEX_S_NOUNDERLINE, INDEX_S_ONELINE, INDEX_S_PASSWORD, INDEX_TOP, INDEX_WIDTH, };

    static const s_property s_propertyTable[];

    void ObjectClassConstructor(char *buff);
    void ObjectDraw(TCanvas *canvas,int left,int top,int width,int height);
    void Set(unsigned int propertyIndex,long value);

  public:

    CObjTextEdit(void);

};
//---------------------------------------------------------------------------
class CObjScrollBar : public CBaseObject
{

  private:

    enum { INDEX_CLASS, INDEX_HORIZONTAL, INDEX_LEFT, INDEX_NAME, INDEX_SIZE, INDEX_TOP, };
    // for size value (which can be width or height) we define phantom values for internal use and catch access...
    enum { INDEX_HEIGHT=900, INDEX_WIDTH, };

    static const s_property s_propertyTable[];

    void ObjectClassConstructor(char *buff);
    void ObjectDraw(TCanvas *canvas,int left,int top,int width,int height);

    long GetValue(unsigned int propertyIndex);
    void Set(unsigned int propertyIndex,long value);
    void GetBounds(int *left,int *top,int *width,int *height);
    void SetBounds(long left,long top,long width,long height);

  public:

    CObjScrollBar(void);

};
//---------------------------------------------------------------------------
class CObjProgressBar : public CBaseObject
{

  private:

    enum { INDEX_CLASS, INDEX_HORIZONTAL, INDEX_LEFT, INDEX_NAME, INDEX_SIZE, INDEX_TOP, };
    // for size value (which can be width or height) we define phantom values for internal use and catch access...
    enum { INDEX_HEIGHT=900, INDEX_WIDTH, };

    static const s_property s_propertyTable[];

    void ObjectClassConstructor(char *buff);
    void ObjectDraw(TCanvas *canvas,int left,int top,int width,int height);

    long GetValue(unsigned int propertyIndex);
    void Set(unsigned int propertyIndex,long value);
    void GetBounds(int *left,int *top,int *width,int *height);
    void SetBounds(long left,long top,long width,long height);

  public:

    CObjProgressBar(void);

};
//---------------------------------------------------------------------------
class CObjLabel : public CBaseObject
{

  private:

    enum { INDEX_CAPTION, INDEX_CLASS, INDEX_HEIGHT, INDEX_LEFT, INDEX_NAME, INDEX_TOP, INDEX_WIDTH, };

    static const s_property s_propertyTable[];

    void ObjectClassConstructor(char *buff);
    void ObjectDraw(TCanvas *canvas,int left,int top,int width,int height);

  public:

    CObjLabel(void);

};
//---------------------------------------------------------------------------
class CObjRect : public CBaseObject
{

  private:

    enum { INDEX_CLASS, INDEX_HEIGHT, INDEX_LEFT, INDEX_NAME, INDEX_TOP, INDEX_WIDTH, };

    static const s_property s_propertyTable[];

    void ObjectClassConstructor(char *buff);
    void ObjectDraw(TCanvas *canvas,int left,int top,int width,int height);

  public:

    CObjRect(void);

};
//---------------------------------------------------------------------------
class CObjIcon : public CBaseObject
{

  private:

    enum { INDEX_CLASS, INDEX_HEIGHT, INDEX_IMAGENAME, INDEX_LEFT, INDEX_NAME, INDEX_TOP, INDEX_WIDTH, };

    int  m_planes;

    static const s_property s_propertyTable[];

    TImage *m_image;

    void ObjectClassConstructor(char *buff);
    void ObjectDraw(TCanvas *canvas,int left,int top,int width,int height);

  public:

    CObjIcon(void);
    ~CObjIcon(void);
    void SetBounds(long left,long top,long width,long height);

};
//---------------------------------------------------------------------------
class CObjTable : public CBaseObject
{

  private:

    enum { INDEX_CLASS, INDEX_HEIGHT, INDEX_LEFT, INDEX_NAME, INDEX_TOP, INDEX_WIDTH, };

    static const s_property s_propertyTable[];

    void ObjectClassConstructor(char *buff);
    void ObjectDraw(TCanvas *canvas,int left,int top,int width,int height);

  public:

    CObjTable(void);

};
//---------------------------------------------------------------------------
class CObjDateEdit : public CBaseObject
{

  private:

    enum { INDEX_CLASS, INDEX_HEIGHT, INDEX_LEFT, INDEX_NAME, INDEX_TOP, INDEX_WIDTH, };

    static const s_property s_propertyTable[];

    void ObjectClassConstructor(char *buff);
    void ObjectDraw(TCanvas *canvas,int left,int top,int width,int height);

  public:

    CObjDateEdit(void);

};
//---------------------------------------------------------------------------
class CObjTimeEdit : public CBaseObject
{

  private:

    enum { INDEX_CLASS, INDEX_HEIGHT, INDEX_LEFT, INDEX_NAME, INDEX_TOP, INDEX_WIDTH, };

    static const s_property s_propertyTable[];

    void ObjectClassConstructor(char *buff);
    void ObjectDraw(TCanvas *canvas,int left,int top,int width,int height);

  public:

    CObjTimeEdit(void);

};
//---------------------------------------------------------------------------
class CObjForm : public CBaseObject
{

  private:

    static const s_property s_propertyTable[];

    void ObjectClassConstructor(char *buff);
    void ObjectDraw(TCanvas *canvas,int left,int top,int width,int height);
    void CObjForm::FormDraw(TCanvas *canvas,int left,int top,int width,int height,bool drawClient);

  public:

    enum
    {
      INDEX_CAPTION, INDEX_CLASS, INDEX_CLOSEBUTTON, INDEX_ENUM, INDEX_HEIGHT, INDEX_LEFT, INDEX_NAME, INDEX_PORTRAIT_O, INDEX_SAVEUNDER, INDEX_TITLEBAR, INDEX_TOP, INDEX_VISIBLE, INDEX_WIDTH,
    };

    CObjForm(void);
    void NonClientDraw(TCanvas *canvas,int offsetX,int offsetY);

};
//---------------------------------------------------------------------------
#endif
